import copy
import math
from typing import Optional, Tuple

import torch
import torch.nn as nn
from torch.nn.utils import spectral_norm as pt_spectral_norm
from transformers.activations import get_activation

from .spectral_norm import spectral_norm


class ElectraClassificationHeadCustom(nn.Module):
    """Head for sentence-level classification tasks."""

    def __init__(self, other):
        super().__init__()
        self.dropout1 = other.dropout
        self.dense = other.dense
        self.dropout2 = copy.deepcopy(other.dropout)
        self.out_proj = other.out_proj

    def forward(self, features, **kwargs):
        x = features[:, 0, :]  # take <s> token (equiv. to [CLS])
        x = self.dropout1(x)
        x = self.dense(x)
        x = get_activation("gelu")(
            x
        )  # although BERT uses tanh here, it seems Electra authors used gelu here
        x = self.dropout2(x)
        x = self.out_proj(x)
        return x


class ElectraClassificationHeadSN(nn.Module):
    """Head for sentence-level classification tasks."""

    def __init__(self, other, sn_value=None, n_power_iterations=1):
        super().__init__()
        self.dropout1 = other.dropout
        if sn_value is None:
            self.dense = pt_spectral_norm(
                other.dense, n_power_iterations=n_power_iterations
            )
        else:
            self.dense = spectral_norm(other.dense, sn_value=sn_value)
        self.dropout2 = copy.deepcopy(other.dropout)
        self.out_proj = other.out_proj

    def eval_init(self, other):
        """This function is used for loading model from saved checkpoint, then we have to load CustomHead too"""
        # Here we load weights of already initialized model, so we don't do spectral_norm
        # also add some weights, that changed after spectral_norm call
        self.dense.weight_orig.data = other["classifier.dense.weight_orig"].data
        self.dense.weight_u.data = other["classifier.dense.weight_u"].data
        self.dense.weight_v.data = other["classifier.dense.weight_v"].data
        self.dense.bias.data = other["classifier.dense.bias"].data
        self.out_proj.weight.data = other["classifier.out_proj.weight"].data
        self.out_proj.bias.data = other["classifier.out_proj.bias"].data

    def forward(self, features, **kwargs):
        x = features[:, 0, :]  # take <s> token (equiv. to [CLS])
        x = self.dropout1(x)
        x = self.dense(x)
        x = get_activation("gelu")(
            x
        )  # although BERT uses tanh here, it seems Electra authors used gelu here
        x = self.dropout2(x)
        x = self.out_proj(x)
        return x


class ElectraNERHeadSN(nn.Module):
    """Head for token-level classification tasks."""

    def __init__(self, other, sn_value=None, n_power_iterations=1):
        super().__init__()
        if sn_value is None:
            self.linear = pt_spectral_norm(
                nn.Linear(768, 768), n_power_iterations=n_power_iterations
            )
        else:
            self.linear = spectral_norm(nn.Linear(768, 768), sn_value=sn_value)
        self.dropout1 = copy.deepcopy(other.dropout)
        self.dropout2 = copy.deepcopy(other.dropout)
        self.out_proj = other.classifier

    def eval_init(self, other):
        """This function is used for loading model from saved checkpoint, then we have to load CustomHead too"""
        # Here we load weights of already initialized model, so we don't do spectral_norm
        # also add some weights, that changed after spectral_norm call
        self.linear.weight_orig.data = other["classifier.linear.weight_orig"].data
        self.linear.weight_u.data = other["classifier.linear.weight_u"].data
        self.linear.weight_v.data = other["classifier.linear.weight_v"].data
        self.linear.bias.data = other["classifier.linear.bias"].data
        self.out_proj.weight.data = other["classifier.out_proj.weight"].data
        self.out_proj.bias.data = other["classifier.out_proj.bias"].data

    def forward(self, features, **kwargs):
        x = features[:, :, :]  # take all
        x = self.dropout1(x)
        x = self.linear(x)
        x = get_activation("gelu")(
            x
        )  # although BERT uses tanh here, it seems Electra authors used gelu here
        x = self.dropout2(x)
        x = self.out_proj(x)
        return x


class BertClassificationHeadIdentityPooler(nn.Module):
    """Head for sentence-level classification tasks."""

    def __init__(self, other):
        super().__init__()

    def forward(self, pooled_output):
        return pooled_output  # , torch.rand((pooled_output.shape[0], pooled_output.shape[0]))


class ElectraClassificationHeadIdentityPooler(nn.Module):
    """Head for sentence-level classification tasks."""

    def __init__(self, other):
        super().__init__()
        self.dropout1 = other.dropout1
        self.dense = other.dense
        # self.dropout2 = copy.deepcopy(other.dropout)
        # self.out_proj = other.out_proj

    def forward(self, features):
        x = features[:, 0, :]  # take <s> token (equiv. to [CLS])
        x = self.dropout1(x)
        x = self.dense(x)
        x = get_activation("gelu")(x)
        return x  # , torch.rand((pooled_output.shape[0], pooled_output.shape[0]))


class XLNetClassificationHeadIdentityPooler(nn.Module):
    """Head for sentence-level classification tasks."""

    def __init__(self):
        super().__init__()

    def forward(self, features):
        return features


class ElectraNERHeadCustom(nn.Module):
    """Head for token-level classification tasks."""

    def __init__(self, other):
        super().__init__()
        self.linear = nn.Linear(768, 768)
        self.dropout1 = copy.deepcopy(other.dropout)
        self.dropout2 = copy.deepcopy(other.dropout)
        self.out_proj = other.classifier

    def eval_init(self, other):
        """This function is used for loading model from saved checkpoint, then we have to load CustomHead too"""
        try:
            if "classifier.out_proj.linear.weight" in other.keys():
                self.linear.weight.data = other[
                    "classifier.out_proj.linear.weight"
                ].data
            else:
                self.linear.weight.data = other[
                    "classifier.out_proj.linear.weight_orig"
                ].data
            self.linear.bias.data = other["classifier.out_proj.linear.bias"].data
            self.out_proj.weight.data = other[
                "classifier.out_proj.out_proj.weight"
            ].data
            self.out_proj.bias.data = other["classifier.out_proj.out_proj.bias"].data
        except:
            if "classifier.linear.weight" in other.keys():
                self.linear.weight.data = other["classifier.linear.weight"].data
            else:
                self.linear.weight.data = other["classifier.linear.weight_orig"].data
            self.linear.bias.data = other["classifier.linear.bias"].data
            self.out_proj.weight.data = other["classifier.out_proj.weight"].data
            self.out_proj.bias.data = other["classifier.out_proj.bias"].data

    def forward(self, features, **kwargs):
        x = features[:, :, :]  # take all
        x = self.dropout1(x)
        x = self.linear(x)
        x = get_activation("gelu")(
            x
        )  # although BERT uses tanh here, it seems Electra authors used gelu here
        x = self.dropout2(x)
        x = self.out_proj(x)
        return x


class ElectraNERHeadIdentityPooler(nn.Module):
    """Head for sentence-level classification tasks."""

    def __init__(self, other):
        super().__init__()
        self.linear = other.linear
        # self.dropout2 = copy.deepcopy(other.dropout)
        # self.out_proj = other.out_proj

    def forward(self, features):
        x = features[:, :, :]  # take all
        x = self.linear(x)
        x = get_activation("gelu")(x)
        return x


def spectral_normalized_model(
    model: torch.nn.Module, substitution_layer_names=["ElectraOutput"]
):
    for i, layer in enumerate(list(model.children())):
        module_name = list(model._modules.items())[i][0]
        layer_name = layer._get_name()
        if layer_name in substitution_layer_names:
            model._modules[module_name].dense = spectral_norm(
                model._modules[module_name].dense
            )
        else:
            spectral_normalized_model(model=layer)


class SpectralNormalizedPooler(torch.nn.Module):
    def __init__(self, pooler, sn_value=None):
        super().__init__()
        if sn_value is None:
            self.dense = pt_spectral_norm(pooler.dense)
        else:
            self.dense = spectral_norm(pooler.dense, sn_value)

        self.dropout = pooler.dropout
        self.config = pooler.config

    def eval_init(self, other):
        """This function is used for loading model from saved checkpoint, then we have to load CustomHead too"""
        # Here we load weights of already initialized model, so we don't do spectral_norm
        # also add some weights, that changed after spectral_norm call
        self.dense.weight_orig.data = other["pooler.dense.weight_orig"].data
        self.dense.weight_u.data = other["pooler.dense.weight_u"].data
        self.dense.weight_v.data = other["pooler.dense.weight_v"].data
        self.dense.bias.data = other["pooler.dense.bias"].data

    def forward(self, hidden_states):
        # We "pool" the model by simply taking the hidden state corresponding
        # to the first token.
        context_token = hidden_states[:, 0]
        context_token = self.dropout(context_token)
        pooled_output = self.dense(context_token)
        pooled_output = get_activation(self.config.pooler_hidden_act)(pooled_output)
        return pooled_output


class ElectraSelfAttentionStochastic(nn.Module):
    def __init__(
        self,
        config,
        position_embedding_type=None,
        hierarchial=False,
        num_centroids=16,
        tau_1=1,
        tau_2=1,
    ):
        super().__init__()
        if config.hidden_size % config.num_attention_heads != 0 and not hasattr(
            config, "embedding_size"
        ):
            raise ValueError(
                f"The hidden size ({config.hidden_size}) is not a multiple of the number of attention "
                f"heads ({config.num_attention_heads})"
            )

        self.num_attention_heads = config.num_attention_heads
        self.attention_head_size = int(config.hidden_size / config.num_attention_heads)
        self.all_head_size = self.num_attention_heads * self.attention_head_size

        self.query = nn.Linear(config.hidden_size, self.all_head_size)
        self.key = nn.Linear(config.hidden_size, self.all_head_size)
        self.value = nn.Linear(config.hidden_size, self.all_head_size)
        self.hierarchial = hierarchial
        self.tau_1 = tau_1
        self.tau_2 = tau_2
        self.num_centroids = num_centroids
        if self.hierarchial:
            self.centroids = nn.Linear(
                self.attention_head_size, self.num_centroids, bias=False
            )

        self.dropout = nn.Dropout(config.attention_probs_dropout_prob)
        self.position_embedding_type = position_embedding_type or getattr(
            config, "position_embedding_type", "absolute"
        )
        if (
            self.position_embedding_type == "relative_key"
            or self.position_embedding_type == "relative_key_query"
        ):
            self.max_position_embeddings = config.max_position_embeddings
            self.distance_embedding = nn.Embedding(
                2 * config.max_position_embeddings - 1, self.attention_head_size
            )

        self.is_decoder = config.is_decoder

    def post_init(self, other):
        # copy weights of other
        if (
            isinstance(other, ElectraSelfAttentionStochastic)
            and self.hierarchial
            and other.hierarchial
        ):
            # copy centroids
            self.centroids.weight.data = other.centroids.weight.data
        for layer in ["query", "key", "value"]:
            getattr(self, layer).weight.data = getattr(other, layer).weight.data
            getattr(self, layer).bias.data = getattr(other, layer).bias.data

    def transpose_for_scores(self, x):
        new_x_shape = x.size()[:-1] + (
            self.num_attention_heads,
            self.attention_head_size,
        )
        x = x.view(new_x_shape)
        return x.permute(0, 2, 1, 3)

    @staticmethod
    def calc_stochastic_attention(attention_scores, tau=1):
        # draw samples from distribution
        samples = nn.functional.gumbel_softmax(attention_scores)
        # calc softmax
        probas = nn.functional.softmax((attention_scores + samples) / tau, dim=-1)
        return probas

    def forward(
        self,
        hidden_states: torch.Tensor,
        attention_mask: Optional[torch.FloatTensor] = None,
        head_mask: Optional[torch.FloatTensor] = None,
        encoder_hidden_states: Optional[torch.FloatTensor] = None,
        encoder_attention_mask: Optional[torch.FloatTensor] = None,
        past_key_value: Optional[Tuple[Tuple[torch.FloatTensor]]] = None,
        output_attentions: Optional[bool] = False,
    ) -> Tuple[torch.Tensor]:
        mixed_query_layer = self.query(hidden_states)

        # If this is instantiated as a cross-attention module, the keys
        # and values come from an encoder; the attention mask needs to be
        # such that the encoder's padding tokens are not attended to.
        is_cross_attention = encoder_hidden_states is not None

        if is_cross_attention and past_key_value is not None:
            # reuse k,v, cross_attentions
            key_layer = past_key_value[0]
            value_layer = past_key_value[1]
            attention_mask = encoder_attention_mask
        elif is_cross_attention:
            key_layer = self.transpose_for_scores(self.key(encoder_hidden_states))
            value_layer = self.transpose_for_scores(self.value(encoder_hidden_states))
            attention_mask = encoder_attention_mask
        elif past_key_value is not None:
            key_layer = self.transpose_for_scores(self.key(hidden_states))
            value_layer = self.transpose_for_scores(self.value(hidden_states))
            key_layer = torch.cat([past_key_value[0], key_layer], dim=2)
            value_layer = torch.cat([past_key_value[1], value_layer], dim=2)
        else:
            key_layer = self.transpose_for_scores(self.key(hidden_states))
            value_layer = self.transpose_for_scores(self.value(hidden_states))

        query_layer = self.transpose_for_scores(mixed_query_layer)

        if self.is_decoder:
            # if cross_attention save Tuple(torch.Tensor, torch.Tensor) of all cross attention key/value_states.
            # Further calls to cross_attention layer can then reuse all cross-attention
            # key/value_states (first "if" case)
            # if uni-directional self-attention (decoder) save Tuple(torch.Tensor, torch.Tensor) of
            # all previous decoder key/value_states. Further calls to uni-directional self-attention
            # can concat previous decoder key/value_states to current projected key/value_states (third "elif" case)
            # if encoder bi-directional self-attention `past_key_value` is always `None`
            past_key_value = (key_layer, value_layer)

        # Take the dot product between "query" and "key" to get the raw attention scores.
        if self.hierarchial:
            # calc a_c with gumbel-softmax
            attention_centroids = self.calc_stochastic_attention(
                self.centroids(key_layer), self.tau_1
            )
            key_centroids = torch.matmul(attention_centroids, self.centroids.weight)
            attention_scores = self.calc_stochastic_attention(
                torch.matmul(query_layer, key_centroids.transpose(-1, -2)), self.tau_2
            )
        else:
            attention_scores = torch.matmul(query_layer, key_layer.transpose(-1, -2))

        if (
            self.position_embedding_type == "relative_key"
            or self.position_embedding_type == "relative_key_query"
        ):
            seq_length = hidden_states.size()[1]
            position_ids_l = torch.arange(
                seq_length, dtype=torch.long, device=hidden_states.device
            ).view(-1, 1)
            position_ids_r = torch.arange(
                seq_length, dtype=torch.long, device=hidden_states.device
            ).view(1, -1)
            distance = position_ids_l - position_ids_r
            positional_embedding = self.distance_embedding(
                distance + self.max_position_embeddings - 1
            )
            positional_embedding = positional_embedding.to(
                dtype=query_layer.dtype
            )  # fp16 compatibility

            if self.position_embedding_type == "relative_key":
                relative_position_scores = torch.einsum(
                    "bhld,lrd->bhlr", query_layer, positional_embedding
                )
                attention_scores = attention_scores + relative_position_scores
            elif self.position_embedding_type == "relative_key_query":
                relative_position_scores_query = torch.einsum(
                    "bhld,lrd->bhlr", query_layer, positional_embedding
                )
                relative_position_scores_key = torch.einsum(
                    "bhrd,lrd->bhlr", key_layer, positional_embedding
                )
                attention_scores = (
                    attention_scores
                    + relative_position_scores_query
                    + relative_position_scores_key
                )

        attention_scores = attention_scores / math.sqrt(self.attention_head_size)
        if attention_mask is not None:
            # Apply the attention mask is (precomputed for all layers in ElectraModel forward() function)
            attention_scores = attention_scores + attention_mask

        # Normalize the attention scores to probabilities.
        # attention_probs = nn.functional.softmax(attention_scores, dim=-1)
        # stochastic modification
        if not self.hierarchial:
            # TODO:
            # mb move before position embeddings
            attention_probs = self.calc_stochastic_attention(
                attention_scores, self.tau_1
            )
        else:
            attention_probs = nn.functional.softmax(attention_scores, dim=-1)

        # This is actually dropping out entire tokens to attend to, which might
        # seem a bit unusual, but is taken from the original Transformer paper.
        attention_probs = self.dropout(attention_probs)

        # Mask heads if we want to
        if head_mask is not None:
            attention_probs = attention_probs * head_mask

        context_layer = torch.matmul(attention_probs, value_layer)

        context_layer = context_layer.permute(0, 2, 1, 3).contiguous()
        new_context_layer_shape = context_layer.size()[:-2] + (self.all_head_size,)
        context_layer = context_layer.view(new_context_layer_shape)

        outputs = (
            (context_layer, attention_probs) if output_attentions else (context_layer,)
        )

        if self.is_decoder:
            outputs = outputs + (past_key_value,)
        return outputs


def replace_attention(model, ue_args, layer_idx):
    model_config = model.electra.encoder.config
    position_embedding_type = model.electra.encoder.layer[
        layer_idx
    ].attention.self.position_embedding_type
    new_attention = ElectraSelfAttentionStochastic(
        model_config,
        position_embedding_type,
        ue_args.hierarchial,
        ue_args.num_centroids,
        ue_args.tau_1,
        ue_args.tau_2,
    )
    new_attention.post_init(model.electra.encoder.layer[layer_idx].attention.self)
    model.electra.encoder.layer[layer_idx].attention.self = new_attention
    return model


def change_attention_params(model, ue_args, layers="all"):
    if layers == "all":
        for layer_idx, _ in enumerate(model.electra.encoder.layer):
            model.electra.encoder.layer[
                layer_idx
            ].attention.self.hierarchial = ue_args.hierarchial
            model.electra.encoder.layer[
                layer_idx
            ].attention.self.num_centroids = ue_args.num_centroids
            model.electra.encoder.layer[layer_idx].attention.self.tau_1 = ue_args.tau_1
            model.electra.encoder.layer[layer_idx].attention.self.tau_2 = ue_args.tau_2
    else:
        layer_idx = -1
        model.electra.encoder.layer[
            layer_idx
        ].attention.self.hierarchial = ue_args.hierarchial
        model.electra.encoder.layer[
            layer_idx
        ].attention.self.num_centroids = ue_args.num_centroids
        model.electra.encoder.layer[layer_idx].attention.self.tau_1 = ue_args.tau_1
        model.electra.encoder.layer[layer_idx].attention.self.tau_2 = ue_args.tau_2
    return model


class ElectraClassificationHS(nn.Module):
    """Head for sentence-level classification tasks."""

    def __init__(self, other, n_labels, hs_labels=None):
        super().__init__()

        self.dropout1 = other.dropout
        self.dense = other.dense

        if hs_labels is None:
            self.H = other.dense.in_features
            self.hs_labels = torch.randn((n_labels, self.H))

            def HS_cost_func(X):
                Z = X @ X.T - 2 * torch.eye(n_labels)
                return Z.max(axis=1).values.mean()

            # L-BFGS
            self.hs_labels.requires_grad = True

            # optimizer = torch.optim.LBFGS([self.hs_labels],
            #                                history_size=100,
            #                                max_iter=200,
            #                                line_search_fn="strong_wolfe")

            optimizer = torch.optim.SGD([self.hs_labels], lr=2e-1)
            scheduler = torch.optim.lr_scheduler.StepLR(
                optimizer, step_size=50, gamma=0.9
            )
            eps = 1e-7
            losses = []

            for i in range(5000):
                optimizer.zero_grad()
                objective = HS_cost_func(self.hs_labels)
                objective.backward()
                optimizer.step(lambda: HS_cost_func(self.hs_labels))
                scheduler.step()
                losses.append(HS_cost_func(self.hs_labels).detach().numpy())
                if len(losses) > 2 and (np.abs(losses[-2] - losses[-1]) < eps):
                    break

            self.hs_labels.requires_grad = False
        else:
            self.hs_labels = hs_labels

    def forward(self, features, **kwargs):
        x = features[:, 0, :]  # take <s> token (equiv. to [CLS])
        x = self.dropout1(x)
        x = self.dense(x)
        self.hs_labels = self.hs_labels.to(x.device)
        x = torch.matmul(x, self.hs_labels.T) * torch.norm(self.hs_labels)
        return x


class BERTClassificationHS(nn.Module):
    """Head for sentence-level classification tasks."""

    def __init__(self, other, n_labels, hs_labels=None):
        super().__init__()
        if hs_labels is None:
            self.H = other.in_features
            self.hs_labels = torch.randn((n_labels, self.H))

            def HS_cost_func(X):
                Z = X @ X.T - 2 * torch.eye(n_labels)
                return Z.max(axis=1).values.mean()

            # L-BFGS
            self.hs_labels.requires_grad = True

            optimizer = torch.optim.LBFGS(
                [self.hs_labels],
                history_size=10,
                max_iter=4,
                line_search_fn="strong_wolfe",
            )
            for i in range(5):
                optimizer.zero_grad()
                objective = HS_cost_func(self.hs_labels)
                objective.backward()
                optimizer.step(lambda: HS_cost_func(self.hs_labels))

            self.hs_labels.requires_grad = False
        else:
            self.hs_labels = hs_labels

    def forward(self, x, **kwargs):
        self.hs_labels = self.hs_labels.to(x.device)
        x = torch.matmul(x, self.hs_labels.T) * torch.norm(self.hs_labels)
        return x



from copy import deepcopy
from typing import Union

import numpy as np
import torch
import torch.nn.functional as F
from datasets import Dataset as ArrowDataset
from torch.utils.data import DataLoader
from tqdm.notebook import tqdm
from transformers import DataCollatorWithPadding

from .model_modifications import (
    get_model_without_cls_layer,
    _get_dim,
    ModelForFeaturesExtraction,
)
from .transformers_dataset import TransformersDataset
from ..model_wrappers.transformers import WrapperCls, WrapperNer, WrapperAts, WrapperNmt

WRAPPERS = {"cls": WrapperCls, "ner": WrapperNer, "ats": WrapperAts, "nmt": WrapperNmt}


def get_embeddings(
    model,
    dataloader_or_data: Union[DataLoader, ArrowDataset, TransformersDataset],
    prepare_model: bool = True,
    use_activation: bool = False,
    use_spectralnorm: bool = False,
    to_eval_mode: bool = True,
    to_numpy: bool = False,
    data_is_tokenized: bool = False,
    batch_size: int = 100,
    use_automodel: bool = False,
    use_averaging: bool = False,
    **tokenization_kwargs,
):
    if prepare_model:
        use_pooling = tokenization_kwargs.get("task", "cls") != "ner"
        model_without_cls_layer = get_model_without_cls_layer(
            model, use_activation, use_spectralnorm, use_pooling=use_pooling
        )
    else:
        model_without_cls_layer = model

    device = next(model_without_cls_layer.parameters()).device

    wrapper = WRAPPERS[tokenization_kwargs.pop("task", "cls")]
    if not isinstance(dataloader_or_data, DataLoader):
        if not data_is_tokenized:
            if "label_name" not in tokenization_kwargs:
                tokenization_kwargs["test_mode"] = True
            tokenize_function = wrapper.get_tokenize_function(
                data=dataloader_or_data,
                **tokenization_kwargs,  # WTF: This should not be binded to the particular Wrapper
            )
            columns_to_remove = [
                x
                for x in dataloader_or_data.features.keys()
                if x not in ["labels", "weight"]
            ]
            # The last two arguments is a temporary fix for Kristofari
            dataloader_or_data = dataloader_or_data.map(
                tokenize_function,
                batched=True,
                remove_columns=columns_to_remove,
                load_from_cache_file=False,
                cache_file_name=f"tmp.data",
            )
            if wrapper == WrapperNer:
                idx_first_bpe = dataloader_or_data["idx_first_bpe"]
                dataloader_or_data = dataloader_or_data.remove_columns("idx_first_bpe")
        dataloader_or_data = DataLoader(
            dataloader_or_data,
            shuffle=False,
            batch_size=batch_size,
            collate_fn=wrapper.get_data_collator_class()(
                tokenizer=tokenization_kwargs["tokenizer"]
            ),
            pin_memory=(str(device).startswith("cuda")),
        )

    num_obs = len(dataloader_or_data.dataset)
    dim = _get_dim(model_without_cls_layer)
    if to_eval_mode:
        model_without_cls_layer.eval()

    embeddings = torch.empty((num_obs, dim), dtype=torch.float, device=device)
    if isinstance(model_without_cls_layer, ModelForFeaturesExtraction):
        possible_input_keys = model_without_cls_layer[
            0
        ].model.forward.__code__.co_varnames
    else:
        possible_input_keys = model_without_cls_layer.forward.__code__.co_varnames
    possible_input_keys = list(possible_input_keys) + ["input_ids", "attention_mask"]

    with torch.no_grad():
        start = 0
        for batch in tqdm(dataloader_or_data, desc="Embeddings created"):
            batch = {
                k: v.to(device) for k, v in batch.items() if k in possible_input_keys
            }
            predictions = model_without_cls_layer(**batch)
            if isinstance(model_without_cls_layer, ModelForFeaturesExtraction):
                batch_embeddings = predictions
            elif use_automodel and not use_activation:
                if use_averaging:
                    batch_embeddings = predictions.last_hidden_state
                    new_embeddings = []
                    for i, emb in enumerate(batch_embeddings):
                        # emb: shape seq_len x dim
                        non_padding_idx = batch["attention_mask"][i].bool()
                        new_embeddings.append(emb[non_padding_idx].mean(dim=0))
                    batch_embeddings = torch.stack(new_embeddings, dim=0)
                else:
                    batch_embeddings = predictions.last_hidden_state[:, 0]
            elif "pooler_output" in predictions.keys():
                batch_embeddings = predictions.pooler_output
            elif "last_hidden_state" in predictions.keys():
                batch_embeddings = predictions.last_hidden_state
            else:
                raise NotImplementedError

            end = start + len(batch["input_ids"])  # len(batch[list(batch.keys())[0]])
            # In this case we need to average batch_embeddings across idx_first_bpe
            if wrapper == WrapperNer:
                new_embeddings = []
                for i, emb in enumerate(batch_embeddings, start=start):
                    # emb: shape seq_len x dim
                    emb = emb[idx_first_bpe[i]].mean(dim=0)
                    new_embeddings.append(emb)
                batch_embeddings = torch.stack(new_embeddings, dim=0)
            embeddings[start:end].copy_(batch_embeddings, non_blocking=True)
            start = end

    if to_numpy:
        return embeddings.cpu().detach().numpy()
    return embeddings


def get_exp_grad_embedding(
    model,
    dataloader_or_data: Union[DataLoader, ArrowDataset, TransformersDataset],
    use_activation: bool = False,
    use_spectralnorm: bool = False,
    to_numpy: bool = False,
    data_is_tokenized: bool = False,
    batch_size: int = 100,
    **tokenization_kwargs,
):
    model_without_cls_layer = get_model_without_cls_layer(
        model, use_activation, use_spectralnorm
    )

    embDim = _get_dim(model_without_cls_layer)

    nLab = model.num_labels

    device = next(model.parameters()).device

    if not isinstance(dataloader_or_data, DataLoader):
        if not data_is_tokenized:
            if "label_name" not in tokenization_kwargs:
                tokenization_kwargs["test_mode"] = True
            tokenize_function = WRAPPERS[
                tokenization_kwargs.pop("task", "cls")
            ].get_tokenize_function(data=dataloader_or_data, **tokenization_kwargs,)
            columns_to_remove = [
                x
                for x in dataloader_or_data.features.keys()
                if x not in ["labels", "weight"]
            ]
            # The last two arguments is a temporary fix for Kristofari
            dataloader_or_data = dataloader_or_data.map(
                tokenize_function,
                batched=True,
                remove_columns=columns_to_remove,
                load_from_cache_file=False,
                cache_file_name=f"tmp.data",
            )

        dataloader_or_data = DataLoader(
            dataloader_or_data,
            shuffle=False,
            batch_size=batch_size,
            collate_fn=DataCollatorWithPadding(
                tokenizer=tokenization_kwargs["tokenizer"]
            ),
            pin_memory=(str(device).startswith("cuda")),
        )

    num_obs = len(dataloader_or_data.dataset)

    embedding = np.zeros([num_obs, nLab, embDim * nLab])

    for ind in range(nLab):
        with torch.no_grad():
            start = 0
            for batch in tqdm(dataloader_or_data, desc="Embeddings created"):
                batch = {
                    k: v.to(device)
                    for k, v in batch.items()
                    if k in model.forward.__code__.co_varnames
                }
                tmp = model(**batch, output_hidden_states=True)
                cout, out = tmp["logits"], tmp["hidden_states"]
                out = out[0]
                out = out.data.cpu().numpy()
                out = np.mean(out, axis=1)

                end = start + len(batch["input_ids"])
                idxs = range(start, end)
                start = end
                batchProbs = F.softmax(cout, dim=1).data.cpu().numpy()
                for j in range(batch_size):
                    if j >= len(out):
                        continue
                    for c in range(nLab):
                        if c == ind:
                            embedding[idxs[j]][ind][
                                embDim * c : embDim * (c + 1)
                            ] = deepcopy(out[j]) * (1 - batchProbs[j][c])
                        else:
                            embedding[idxs[j]][ind][
                                embDim * c : embDim * (c + 1)
                            ] = deepcopy(out[j]) * (-1 * batchProbs[j][c])
                    embedding[idxs[j]][ind] = embedding[idxs[j]][ind] * np.sqrt(
                        batchProbs[j][ind]
                    )
    if to_numpy:
        return embedding

    return torch.Tensor(embedding)


def get_grad_embeddings(
    model,
    dataloader_or_data: Union[DataLoader, ArrowDataset, TransformersDataset],
    use_activation: bool = False,
    use_spectralnorm: bool = False,
    to_numpy: bool = False,
    data_is_tokenized: bool = False,
    batch_size: int = 100,
    **tokenization_kwargs,
):
    model_without_cls_layer = get_model_without_cls_layer(
        model, use_activation, use_spectralnorm
    )

    embDim = _get_dim(model_without_cls_layer)

    nLab = model.num_labels

    device = next(model.parameters()).device

    if not isinstance(dataloader_or_data, DataLoader):
        if not data_is_tokenized:
            if "label_name" not in tokenization_kwargs:
                tokenization_kwargs["test_mode"] = True
            tokenize_function = WRAPPERS[
                tokenization_kwargs.pop("task", "cls")
            ].get_tokenize_function(**tokenization_kwargs,)
            columns_to_remove = [
                x
                for x in dataloader_or_data.features.keys()
                if x not in ["labels", "weight"]
            ]
            # The last two arguments is a temporary fix for Kristofari
            dataloader_or_data = dataloader_or_data.map(
                tokenize_function,
                batched=True,
                remove_columns=columns_to_remove,
                load_from_cache_file=False,
                cache_file_name=f"tmp.data",
            )

        dataloader_or_data = DataLoader(
            dataloader_or_data,
            shuffle=False,
            batch_size=batch_size,
            collate_fn=DataCollatorWithPadding(
                tokenizer=tokenization_kwargs["tokenizer"]
            ),
            pin_memory=(str(device).startswith("cuda")),
        )

    num_obs = len(dataloader_or_data.dataset)
    embedding = np.zeros([num_obs, embDim * nLab])
    for ind in range(nLab):
        with torch.no_grad():
            start = 0
            for batch in tqdm(dataloader_or_data, desc="Embeddings created"):
                batch = {
                    k: v.to(device)
                    for k, v in batch.items()
                    if k in model.forward.__code__.co_varnames
                }
                tmp = model(**batch, output_hidden_states=True)
                cout, out = tmp["logits"], tmp["hidden_states"]
                out = out[0]
                out = out.data.cpu().numpy()
                out = np.mean(out, axis=1)

                end = start + len(batch["input_ids"])
                idxs = range(start, end)
                start = end

                batchProbs = F.softmax(cout, dim=1).data.cpu().numpy()
                maxInds = np.argmax(batchProbs, 1)
                for j in range(batch_size):
                    if j >= len(out):
                        continue
                    for c in range(nLab):
                        if c == maxInds[j]:
                            embedding[idxs[j]][
                                embDim * c : embDim * (c + 1)
                            ] = deepcopy(out[j]) * (1 - batchProbs[j][c])
                        else:
                            embedding[idxs[j]][
                                embDim * c : embDim * (c + 1)
                            ] = deepcopy(out[j]) * (-1 * batchProbs[j][c])

    if to_numpy:
        return embedding

    return torch.Tensor(embedding)



from typing import Union

from datasets.arrow_dataset import Dataset

from .al_strategy_utils import (
    take_idx,
    calculate_actune_scores,
)
from ..utils.transformers_dataset import TransformersDataset


def actune(
    model,
    X_pool: Union[Dataset, TransformersDataset],
    n_instances: int,
    query_strategy,
    **actune_kwargs,
):
    _, _, uncertainty_estimates = query_strategy(
        model, X_pool, n_instances, **actune_kwargs
    )

    probas = model.predict_proba(X_pool)
    kwargs = dict(
        # Necessary
        model=model,
        data_test=X_pool,
        # General
        data_is_tokenized=False,
        data_config=model.data_config,
        batch_size=model._batch_size_kwargs.eval_batch_size,
        probas=probas,
        n_instances=n_instances,
        uncertainty_estimates=uncertainty_estimates,
        k=10,
    )

    query_idx, uncertainty_estimates = calculate_actune_scores(**kwargs)
    query = take_idx(X_pool, query_idx)

    return query_idx, query, uncertainty_estimates



from pathlib import Path

import numpy as np
from datasets import concatenate_datasets
from sklearn.model_selection import train_test_split
from torch import tensor
from transformers import AutoTokenizer


def preprocess_img(example):
    data = tensor(np.array(example["image"]))
    data = data.float().div(255)
    data = data.sub_(0.1307).div_(0.3081)
    data = list(data.numpy())
    example["image"] = data
    return example


def _add_id_column_to_datasets(datasets):
    """
    :param datasets: should be strictly in order `train - dev - test (if test exists)`
    :return:
    """
    last_id = 0
    for i in range(len(datasets)):
        dataset = datasets[i]
        if dataset is not None:
            datasets[i] = dataset.add_column(
                "id", list(range(last_id, last_id + len(dataset)))
            )
            last_id += len(dataset)
        return datasets


def _use_train_subset(train_dataset, subset_size, seed, task, label_name):
    kwargs = {"random_state": seed}
    if task.endswith("cls"):
        kwargs["stratify"] = train_dataset[label_name]
    random_train_idx = train_test_split(
        range(len(train_dataset)), train_size=subset_size, **kwargs
    )[0]
    return train_dataset.select(random_train_idx)


def _use_test_subset(test_dataset, subset_share_or_size, seed=42, use_seed_42=False):
    if isinstance(subset_share_or_size, float):
        subset_size = round(subset_share_or_size * len(test_dataset))
    else:
        subset_size = subset_share_or_size
    if use_seed_42:
        np.random.seed(42)
    else:
        np.random.seed(seed)
    subsample_idx = np.random.choice(range(len(test_dataset)), subset_size, False)
    return test_dataset.select(subsample_idx), subsample_idx


def _filter_quantiles(train_dataset, quantiles, cache_dir, text_name, tokenizer_name):
    tokenizer = AutoTokenizer.from_pretrained(
        tokenizer_name,
        cache_dir=Path(cache_dir) / "tokenizer" if cache_dir is not None else None,
    )
    train_lengths = np.array(
        [
            len(x)
            for x in tokenizer(train_dataset[text_name], truncation=False)["input_ids"]
        ]
    )
    length_quantiles = np.quantile(train_lengths, quantiles)
    ids_satisfy_cond = np.argwhere(
        (train_lengths >= length_quantiles[0]) & (train_lengths <= length_quantiles[1])
    ).ravel()
    return train_dataset.select(ids_satisfy_cond)


def _multiply_data(hfdreader, multiply_coef):
    if type(hfdreader.dataset.num_rows) == int:
        hfdreader.dataset = concatenate_datasets([hfdreader.dataset] * multiply_coef)
    else:
        # we have dataset with several splits
        for key in hfdreader.dataset.num_rows.keys():
            if key != "test":
                hfdreader.dataset[key] = concatenate_datasets(
                    [hfdreader.dataset[key]] * multiply_coef
                )
    return hfdreader



from typing import Union

import numpy as np
from datasets import concatenate_datasets
from datasets.arrow_dataset import Dataset

from ..al_strategy_utils import take_idx
from ...utils.transformers_dataset import TransformersDataset


def small_text_sampling(
    model,
    X_pool: Union[Dataset, TransformersDataset],
    n_instances: int,
    X_train: Union[Dataset, TransformersDataset],
    small_text_strategy,
    **kwargs,
):
    from small_text.query_strategies import (
        LeastConfidence,
        PredictionEntropy,
    )

    small_text_strategy_dict = {
        "lc": LeastConfidence,
        "ent": PredictionEntropy,
    }
    all_data = concatenate_datasets([X_train, X_pool])
    indices_labeled = np.arange(len(X_train))
    indices_unlabeled = np.arange(len(X_train), len(X_pool))
    small_text_strategy = small_text_strategy_dict[small_text_strategy]()

    query_idx = small_text_strategy.query(
        model, all_data, indices_unlabeled, indices_labeled, None, n=n_instances,
    )

    uncertainty_estimates = np.zeros(len(X_pool))
    query = take_idx(all_data, query_idx)

    return query_idx, query, uncertainty_estimates



import logging
from pathlib import Path
from typing import Union

import numpy as np
import torch
import yaml
from datasets.arrow_dataset import Dataset
from torch.nn.functional import normalize

from .al_strategy_utils import (
    get_query_idx_for_selecting_by_number_of_tokens,
    calculate_mnlp_score,
    calculate_bald_score_cls,
    calculate_alps_scores,
    take_idx,
    calculate_mahalanobis_distance,
    calculate_mahalanobis_triplet_scores,
    calculate_mahalanobis_filtering_scores,
    calculate_triplet_scores,
    calculate_ddu_scores,
    calculate_ddu_scores_cv,
    choose_cm_samples,
)
from .strategy_utils import get_batchbald_batch
from .strategy_utils import make_dropouts_consistent
from .strategy_utils import UeEstimatorHybrid, TextClassifier
from ..utils.augmentation import augment_instances
from ..utils.cluster_utils import badge, kmeans
from ..utils.transformers_dataset import TransformersDataset

log = logging.getLogger()


def random_sampling(
    model,
    X_pool,
    n_instances,
    *args,
    select_by_number_of_tokens: bool = False,
    **kwargs,
):
    if select_by_number_of_tokens:
        sorted_idx = np.arange(len(X_pool))
        np.random.shuffle(sorted_idx)
        query_idx = get_query_idx_for_selecting_by_number_of_tokens(
            X_pool, sorted_idx, n_instances
        )
    else:
        query_idx = np.random.choice(range(len(X_pool)), n_instances, replace=False)

    query = take_idx(X_pool, query_idx)
    # Uncertainty estimates are not defined with random sampling
    uncertainty_estimates = np.zeros(len(X_pool))
    return query_idx, query, uncertainty_estimates


def ssal_sampling(model, X_pool, n_instances):
    X_pool_augmented = augment_instances(X_pool)
    num_augmentations = int(len(X_pool_augmented) / len(X_pool))
    probas = np.reshape(
        model.predict_proba(X_pool_augmented), [len(X_pool), num_augmentations, -1]
    )
    uncertainty_estimates = probas.var(axis=1).sum(axis=1)
    argsort = np.argsort(-uncertainty_estimates)
    query_idx = argsort[:n_instances]
    query = take_idx(X_pool, query_idx)

    return query_idx, query, uncertainty_estimates


def lc_sampling(
    model,
    X_pool: Union[np.ndarray, Dataset, TransformersDataset],
    n_instances: int,
    **kwargs,
):
    """
    uncertainty estimates: the larger the value, the less confident the model is
    :param classifier:
    :param X_pool:
    :param n_instances:
    :param kwargs:
    :return:
    """
    probas = model.predict_proba(X_pool)
    uncertainty_estimates = 1 - np.max(probas, axis=1)
    argsort = np.argsort(-uncertainty_estimates)
    query_idx = argsort[:n_instances]
    query = take_idx(X_pool, query_idx)

    return query_idx, query, uncertainty_estimates


def entropy_sampling(
    model,
    X_pool: Union[np.ndarray, Dataset, TransformersDataset],
    n_instances: int,
    **kwargs,
):
    """
    uncertainty estimates: the larger the value, the less confident the model is
    :param classifier:
    :param X_pool:
    :param n_instances:
    :param kwargs:
    :return:
    """
    probas = model.predict_proba(X_pool)
    uncertainty_estimates = np.sum(-probas * np.log(probas), axis=1)
    argsort = np.argsort(-uncertainty_estimates)
    query_idx = argsort[:n_instances]
    query = take_idx(X_pool, query_idx)

    return query_idx, query, uncertainty_estimates


def mnlp_sampling(
    model,
    X_pool: Union[np.ndarray, Dataset, TransformersDataset],
    n_instances: int,
    select_by_number_of_tokens=True,
    **kwargs,
):
    probas = model.predict_proba(X_pool, remove_padding=True)
    np.array([-np.sum(np.log(np.max(i, axis=1))) / len(i) for i in probas])

    uncertainty_estimates = calculate_mnlp_score(probas)
    argsort = np.argsort(-uncertainty_estimates)

    if select_by_number_of_tokens:
        query_idx = get_query_idx_for_selecting_by_number_of_tokens(
            X_pool, argsort, n_instances
        )
    else:
        query_idx = argsort[:n_instances]

    query = take_idx(X_pool, query_idx)
    return query_idx, query, uncertainty_estimates


def old_mahalanobis_sampling(
    model, X_pool, n_instances, **mahalanobis_kwargs,
):
    kwargs = dict(
        # Necessary
        model_wrapper=model,
        data=X_pool,
        # General
        data_is_tokenized=False,
        data_config=None,
        batch_size=model._batch_size_kwargs.eval_batch_size,
        to_numpy=True,
        # Mahalanobis
        use_v2=mahalanobis_kwargs["use_v2"],
        use_activation=mahalanobis_kwargs["use_activation"],
        use_spectralnorm=mahalanobis_kwargs["use_spectralnorm"],
    )
    dists = calculate_mahalanobis_distance(**kwargs)
    uncertainty_estimates = np.min(dists, axis=-1)
    argsort = np.argpartition(-uncertainty_estimates, kth=n_instances)
    query_idx = argsort[:n_instances]
    query = take_idx(X_pool, query_idx)

    return query_idx, query, uncertainty_estimates


def mahalanobis_sampling(
    model,
    X_pool: Union[Dataset, TransformersDataset],
    n_instances: int,
    X_train: Union[Dataset, TransformersDataset],
    **mahalanobis_kwargs,
):
    batched = mahalanobis_kwargs.get("mahalanobis_batched", False)
    substrategy = mahalanobis_kwargs.get("mahalanobis_substrategy", "lc")
    use_class_probas = mahalanobis_kwargs.get("mahalanobis_use_class_probas", False)
    kwargs = dict(
        # Necessary
        model_wrapper=model,
        train_data=X_train,
        unlabeled_data=X_pool,
        # General
        data_is_tokenized=False,
        data_config=None,
        batch_size=model._batch_size_kwargs.eval_batch_size,
        to_numpy=True,
        # Mahalanobis
        classwise=True,  # use an own centroid for each class
        batched=batched,
        use_v2=mahalanobis_kwargs.get("mahalanobis_use_v2", False),
        use_activation=mahalanobis_kwargs.get("mahalanobis_use_activation", False),
        use_spectralnorm=mahalanobis_kwargs.get("mahalanobis_use_spectralnorm", True),
    )
    if not batched:
        dists = calculate_mahalanobis_distance(**kwargs)
        if use_class_probas:
            log.info("Using probas for mahalanobis")
            label_name = model.data_config["label_name"]
            class_probas = np.bincount(
                X_train[label_name], minlength=model.num_labels
            ) / len(X_train)
            dists = dists * class_probas
        else:
            log.info("Not using probas for mahalanobis))")
        if substrategy == "lc":
            uncertainty_estimates = np.min(dists, axis=-1)
        elif substrategy == "margin":
            dists.sort(axis=1)
            max_dists = dists[:, 0]
            second_max_dists = dists[:, 1]
            uncertainty_estimates = (
                max_dists - second_max_dists
            )  # the greater the value, the more uncertain we are
        else:
            raise NotImplementedError
        argsort = np.argpartition(-uncertainty_estimates, kth=n_instances)
        query_idx = argsort[:n_instances]
    else:
        query_idx, uncertainty_estimates = calculate_mahalanobis_distance(
            n_instances=n_instances, **kwargs
        )
    query = take_idx(X_pool, query_idx)
    return query_idx, query, uncertainty_estimates


def mahalanobis_triplet_sampling(
    model,
    X_pool: Union[Dataset, TransformersDataset],
    n_instances: int,
    X_train: Union[Dataset, TransformersDataset],
    **mahalanobis_triplet_kwargs,
):
    batched = mahalanobis_triplet_kwargs.get("mahalanobis_batched", False)
    kwargs = dict(
        # Necessary
        model_wrapper=model,
        train_data=X_train,
        unlabeled_data=X_pool,
        # General
        data_is_tokenized=False,
        data_config=None,
        batch_size=model._batch_size_kwargs.eval_batch_size,
        to_numpy=True,
        # Mahalanobis Triplet
        batched=batched,
        lamb=mahalanobis_triplet_kwargs.get("mahalanobis_triplet_lamb", 0.25),
        use_finetuned=mahalanobis_triplet_kwargs.get("mahalanobis_use_finetuned", True),
        use_v2=mahalanobis_triplet_kwargs.get("mahalanobis_use_v2", False),
        use_activation=mahalanobis_triplet_kwargs.get(
            "mahalanobis_use_activation", False
        ),
        use_spectralnorm=mahalanobis_triplet_kwargs.get(
            "mahalanobis_use_spectralnorm", True
        ),
    )
    if not batched:
        uncertainty_estimates = calculate_mahalanobis_triplet_scores(**kwargs)
        argsort = np.argpartition(-uncertainty_estimates, kth=n_instances)
        query_idx = argsort[:n_instances]
    else:
        query_idx, uncertainty_estimates = calculate_mahalanobis_triplet_scores(
            n_instances=n_instances, **kwargs
        )

    query = take_idx(X_pool, query_idx)
    return query_idx, query, uncertainty_estimates


def triplet_sampling(
    model,
    X_pool: Union[Dataset, TransformersDataset],
    n_instances: int,
    X_train: Union[Dataset, TransformersDataset],
    **triplet_kwargs,
):
    kwargs = dict(
        # Necessary
        model_wrapper=model,
        train_data=X_train,
        unlabeled_data=X_pool,
        # General
        data_is_tokenized=False,
        data_config=None,
        batch_size=model._batch_size_kwargs.eval_batch_size,
        # Mahalanobis Triplet
        lamb=triplet_kwargs.get("triplet_lamb", 0.25),
        strategy=triplet_kwargs.get("uncertainty_strategy", "lc"),
        use_finetuned=triplet_kwargs.get("mahalanobis_use_finetuned", True),
        scale_distances=triplet_kwargs.get("scale_distances", True),
        use_activation=triplet_kwargs.get("mahalanobis_use_activation", False),
        use_spectralnorm=triplet_kwargs.get("mahalanobis_use_spectralnorm", True),
    )
    uncertainty_estimates = calculate_triplet_scores(**kwargs)
    argsort = np.argpartition(-uncertainty_estimates, kth=n_instances)
    query_idx = argsort[:n_instances]

    query = take_idx(X_pool, query_idx)
    return query_idx, query, uncertainty_estimates


def mahalanobis_filtering_sampling(
    model,
    X_pool: Union[Dataset, TransformersDataset],
    n_instances: int,
    X_train: Union[Dataset, TransformersDataset],
    **filtering_kwargs,
):
    batched = False
    if "mahalanobis_batched" in filtering_kwargs:
        batched = filtering_kwargs.pop("mahalanobis_batched")
    kwargs = dict(
        # Necessary
        model_wrapper=model,
        train_data=X_train,
        unlabeled_data=X_pool,
        # General
        data_is_tokenized=False,
        data_config=None,
        batch_size=model._batch_size_kwargs.eval_batch_size,
        # Mahalanobis Filtering
        filtering_share=filtering_kwargs.get("filtering_share", 0.01),
        strategy=filtering_kwargs.get("uncertainty_strategy", "lc"),
        batched=batched,
        n_instances=n_instances,
        use_finetuned=filtering_kwargs.get("mahalanobis_use_finetuned", True),
        use_activation=filtering_kwargs.get("mahalanobis_use_activation", False),
        use_spectralnorm=filtering_kwargs.get("mahalanobis_use_spectralnorm", True),
    )
    if batched:
        query_idx, uncertainty_estimates = calculate_mahalanobis_filtering_scores(
            **kwargs
        )
    else:
        uncertainty_estimates = calculate_mahalanobis_filtering_scores(**kwargs)
        argsort = np.argpartition(-uncertainty_estimates, kth=n_instances)
        query_idx = argsort[:n_instances]

    query = take_idx(X_pool, query_idx)
    return query_idx, query, uncertainty_estimates


def badge_sampling(
    model,
    X_pool: Union[Dataset, TransformersDataset],
    n_instances: int,
    X_train: Union[Dataset, TransformersDataset],
    **badge_kwargs,
):
    logits = model.predict_logits(X_pool)

    kwargs = dict(
        # Necessary
        model_wrapper=model,
        data_test=X_pool,
        # General
        data_is_tokenized=False,
        data_config=None,
        batch_size=model._batch_size_kwargs.eval_batch_size,
        to_numpy=True,
        logits=logits,
        # train_probas=train_logits,
    )

    scores_or_vectors = calculate_badge_scores(**kwargs)

    # cluster-based sampling method like BADGE and ALPS
    vectors = normalize(scores_or_vectors)
    # centers = _sampled.tolist()

    clustering = badge

    query_idx = np.array(clustering(vectors, k=n_instances))

    query = take_idx(X_pool, query_idx)

    # TODO
    # Define Uncertainty estimates for BADGE sampling
    uncertainty_estimates = np.zeros(len(X_pool))

    return query_idx, query, uncertainty_estimates


def alps_sampling(
    model,
    X_pool: Union[Dataset, TransformersDataset],
    n_instances: int,
    X_train: Union[Dataset, TransformersDataset],
    **alps_kwargs,
):
    logits = model.predict_logits(X_pool)

    uncertainty_estimates = 1 - np.max(logits, axis=1)
    kwargs = dict(
        # Necessary
        model_wr=model,
        dataloader_or_data=X_pool,
        # General
        data_is_tokenized=False,
        # data_config=None,
        batch_size=model._batch_size_kwargs.eval_batch_size,
        to_numpy=True,
        logits=logits,
        # train_probas=train_logits,
        # data_is_tokenized=data_is_tokenized,
        tokenizer=model.tokenizer,
        task=model.task,
        text_name=model.data_config["text_name"],
        label_name=model.data_config["label_name"],
    )
    scores_or_vectors = calculate_alps_scores(**kwargs)
    # cluster-based sampling method like BADGE and ALPS
    vectors = normalize(scores_or_vectors)
    clustering = kmeans

    query_idx = np.array(
        clustering(
            # vectors[unsampled], k = args.query_size
            vectors,
            k=n_instances,
        )
    )
    # add new samples to previously sampled list
    query = take_idx(X_pool, query_idx)

    return query_idx, query, uncertainty_estimates


def cal_sampling(
    model,
    X_pool: Union[Dataset, TransformersDataset],
    n_instances: int,
    X_train: Union[Dataset, TransformersDataset],
    **cal_kwargs,
):
    logits = model.predict_logits(X_pool)
    train_logits = model.predict_logits(X_train)
    kwargs = dict(
        # Necessary
        model_wrapper=model,
        data_train=X_train,
        data_test=X_pool,
        # General
        data_is_tokenized=False,
        data_config=None,
        batch_size=model._batch_size_kwargs.eval_batch_size,
        to_numpy=True,
        probas=logits,
        train_probas=train_logits,
        num_nei=cal_kwargs["num_nei"],
    )

    uncertainty_estimates = calculate_cal_scores(**kwargs)
    # argsort = np.argsort(-uncertainty_estimates)
    # query_idx = argsort[:n_instances]
    query_idx = np.argpartition(uncertainty_estimates, -n_instances)[-n_instances:]
    query = take_idx(X_pool, query_idx)
    return query_idx, query, uncertainty_estimates


def ddu_sampling(
    model,
    X_pool: Union[Dataset, TransformersDataset],
    n_instances: int,
    X_train: Union[Dataset, TransformersDataset],
    **ddu_kwargs,
):
    kwargs = dict(
        # Necessary
        model_wrapper=model,
        data_train=X_train,
        data_test=X_pool,
        # General
        data_is_tokenized=False,
        data_config=None,
        batch_size=model._batch_size_kwargs.eval_batch_size,
        to_numpy=True,
        # DDU
        use_activation=ddu_kwargs["use_activation"],
        use_spectralnorm=ddu_kwargs["use_spectralnorm"],
    )
    uncertainty_estimates = calculate_ddu_scores(**kwargs)
    argsort = np.argpartition(uncertainty_estimates, kth=n_instances)
    query_idx = argsort[:n_instances]
    query = take_idx(X_pool, query_idx)

    return query_idx, query, -uncertainty_estimates


def logits_lc_sampling(
    model,
    X_pool: Union[np.ndarray, Dataset, TransformersDataset],
    n_instances: int,
    **kwargs,
):
    probas = model.predict_logits(X_pool)
    uncertainty_estimates = 1 - np.max(probas, axis=1)
    argsort = np.argsort(-uncertainty_estimates)
    query_idx = argsort[:n_instances]
    query = take_idx(X_pool, query_idx)

    return query_idx, query, uncertainty_estimates


def margin_sampling(
    model,
    X_pool: Union[np.ndarray, Dataset, TransformersDataset],
    n_instances: int,
    **kwargs,
):
    probas = model.predict_proba(X_pool)
    # To get second max probas, need to sort the array since `.sort` modifies the array
    probas.sort(axis=1)
    max_probas = probas[:, -1]
    second_max_probas = probas[:, -2]
    uncertainty_estimates = 1 + second_max_probas - max_probas
    argsort = np.argsort(-uncertainty_estimates)
    query_idx = argsort[:n_instances]
    query = X_pool.select(query_idx)

    return query_idx, query, uncertainty_estimates


def oracle_sampling(
    model,
    X_pool: Union[np.ndarray, Dataset, TransformersDataset],
    n_instances: int,
    **kwargs,
):
    """
    ¡NB! Works only with `transformers` framework
    """
    probas = model.predict_proba(X_pool)
    labels = X_pool["label"]
    uncertainty_estimates = -np.log([pr[lab] for pr, lab in zip(probas, labels)])
    argsort = np.argsort(-uncertainty_estimates)
    query_idx = argsort[:n_instances]
    query = take_idx(X_pool, query_idx)

    return query_idx, query, uncertainty_estimates


def bald_sampling(
    model,
    X_pool: Union[np.ndarray, Dataset, TransformersDataset],
    n_instances: int,
    select_by_number_of_tokens: bool = False,
    **bald_kwargs,
):
    mc_iterations = bald_kwargs.get("mc_iterations", 10)
    use_stable_dropout = bald_kwargs.get("use_stable_dropout", 10)

    # TODO: realize whether is required or not
    # Make dropout consistent inside huggingface model
    if use_stable_dropout:
        make_dropouts_consistent(model.model)
    else:
        model.enable_dropout()

    if bald_kwargs.get("only_head_dropout", False):
        raise NotImplementedError
        # log_probs_N_K_C = model.predict_proba(X_pool, mc_dropout=True, mc_iterations=mc_iterations)
    else:
        # Stable dropout
        probas = []
        for _ in range(mc_iterations):
            if use_stable_dropout:
                # Reset masks
                model.enable_dropout()
                model.disable_dropout()
            probas_iter = model.predict_proba(
                X_pool, use_predict_loop=True, to_eval_mode=False
            )
            probas.append(probas_iter)

        probas_N_K_C = np.stack(probas, -2)

    uncertainty_estimates = calculate_bald_score_cls(probas_N_K_C)
    # The larger the score, the more confident the model is
    argsort = np.argsort(-uncertainty_estimates)

    if select_by_number_of_tokens:
        query_idx = get_query_idx_for_selecting_by_number_of_tokens(
            X_pool, argsort, n_instances
        )
    else:
        query_idx = argsort[:n_instances]
    query = take_idx(X_pool, query_idx)

    return query_idx, query, uncertainty_estimates


def batchbald_sampling(
    model,
    X_pool: Union[np.ndarray, Dataset, TransformersDataset],
    n_instances: int,
    **bald_kwargs,
):

    mc_iterations = bald_kwargs.get("mc_iterations", 10)
    max_num_samples = bald_kwargs.get("max_num_samples", int(1e4))
    # requires enormous amount of memory
    device = "cpu"  # list(model.model.parameters())[0].device

    # Make dropout consistent inside huggingface model
    make_dropouts_consistent(model.model)

    if bald_kwargs.get("only_head_dropout", False):
        raise NotImplementedError
        # log_probas_N_K_C = model.predict_proba(X_pool, mc_dropout=True, mc_iterations=mc_iterations)
    else:
        probas = []
        for _ in range(mc_iterations):
            # Reset masks
            model.enable_dropout()
            model.disable_dropout()
            probas_iter = model.predict_proba(
                X_pool, use_predict_loop=True, to_numpy=False, to_eval_mode=False
            )
            probas.append(probas_iter)

        log_probas_N_K_C = torch.stack(probas, -2).log().to(device)

    query_idx, uncertainty_estimates = get_batchbald_batch(
        log_probas_N_K_C, n_instances, max_num_samples, device=device
    )
    query = take_idx(X_pool, query_idx)

    return query_idx, query, uncertainty_estimates


def hybrid_sampling(
    model,
    X_pool: Union[np.ndarray, Dataset, TransformersDataset],
    n_instances: int,
    X_train: Union[np.ndarray, Dataset, TransformersDataset],
    **kwargs,
):
    path_to_ue_config = kwargs.get(
        "path_to_ue_config", "/home/jovyan/active_learning/ue/configs/ue/hybrid.yaml"
    )
    if not Path(path_to_ue_config).is_file():
        raise ValueError(f"Check path to UE config! Current: {path_to_ue_config}")

    with open(path_to_ue_config) as f:
        ue_cfg = yaml.load(f, yaml.Loader)

    training_cfg = {
        "learning_rate": model.trainer.args.learning_rate,
        "per_device_train_batch_size": model.trainer.args.per_device_train_batch_size,
        "per_device_eval_batch_size": model.trainer.args.per_device_eval_batch_size,
        "num_train_epochs": model.trainer.args.num_train_epochs,
        "warmup_ratio": model.trainer.args.warmup_ratio,
        "seed": model.trainer.args.seed,
    }

    tokenizer = model.tokenizer
    task = model.task
    text_name = model.data_config["text_name"]
    label_name = model.data_config["label_name"]

    cls = TextClassifier(
        model.model,
        tokenizer,
        max_len=model.tokenizer.model_max_length,
        training_args=training_cfg,
        trainer=model.trainer,
    )

    tokenized_train = model.tokenize_data(tokenizer, X_train, text_name, label_name)
    tokenized_pool = model.tokenize_data(tokenizer, X_pool, text_name, label_name)

    estimator = UeEstimatorHybrid(cls, ue_cfg, None, tokenized_train)
    estimator.fit_ue(tokenized_train)
    uncertainty_estimates = np.array(estimator(tokenized_pool)["mahalanobis_distance"])
    estimator._return_head()
    argsort = np.argsort(-uncertainty_estimates)
    query_idx = argsort[:n_instances]
    query = take_idx(X_pool, query_idx)

    return query_idx, query, uncertainty_estimates


def cluster_margin(classifier, X_pool, n_instances, **kwargs):
    instances_multiplier = kwargs["instances_multiplier"]
    random_query = kwargs["random_query"]
    probas = classifier.predict_proba(X_pool)
    # reuse code from margin sampling
    # To get second max probas, need to sort the array since `.sort` modifies the array
    probas.sort(axis=1)
    max_probas = probas[:, -1]
    second_max_probas = probas[:, -2]
    uncertainty_estimates = 1 + second_max_probas - max_probas
    argsort = np.argsort(-uncertainty_estimates)
    # we have to choose n_instances with round-robin algorithm, so for this we
    # firstly choose 2 * n_instances
    query_idx = argsort[: int(instances_multiplier * n_instances)]
    cluster_labels = np.array(X_pool.clusters)[query_idx]
    # X_pool either transformer dataset or np array
    query_idx, samples_idx = choose_cm_samples(
        cluster_labels, query_idx, n_instances, random_query
    )
    query = take_idx(X_pool, query_idx)
    return query_idx, query, uncertainty_estimates


# def de_ue_cls(
#     classifiers,
#     X_pool,
#     n_instances,
#     gamma_or_k_confident_to_save=0.25,
#     T=None,
#     sampling_type=None,
#     require_sampling=True,
#     method="bald",
#     **kwargs,
# ):
#     prob_arr = np.stack(
#         [classifier.predict_proba(X_pool) for classifier in classifiers], axis=1
#     )
#     if method == "bald":
#         scores = calculate_bald_score_cls(prob_arr)
#     elif method == "vr":
#         scores = var_ratio(prob_arr)
#     elif method == "mean_ent":
#         scores = mean_entropy(prob_arr)
#     elif method == "max_prob":
#         scores = sampled_max_prob(prob_arr)
#     elif method == "prob_var":
#         scores = probability_variance(prob_arr)
#     else:
#         raise NotImplementedError
#
#     argsort = np.argsort(-scores)
#     query_idx = argsort[:n_instances]
#     query = (
#         X_pool[query_idx]
#         if isinstance(X_pool, np.ndarray)
#         else X_pool.select(query_idx)
#     )
#
#     if (sampling_type is None) or (not require_sampling):
#         return query_idx, query
#
#     elif sampling_type == "naive":
#         k_confident_idx = argsort[
#             n_instances : int(gamma_or_k_confident_to_save * len(argsort)) + n_instances
#         ]
#     elif sampling_type == "ups":
#         k_confident_idx = sample_ups(
#             argsort, query_idx, gamma_or_k_confident_to_save, T
#         )
#     elif sampling_type == "random":
#         idx_to_select_from = np.setdiff1d(argsort, query_idx)
#         share_to_sample = gamma_or_k_confident_to_save
#         k_confident_idx = np.random.choice(
#             idx_to_select_from,
#             int(share_to_sample * len(idx_to_select_from)),
#             replace=False,
#         )
#     else:
#         raise NotImplementedError
#
#     k_confident_query = (
#         X_pool[k_confident_idx]
#         if isinstance(X_pool, np.ndarray)
#         else X_pool.select(k_confident_idx)
#     )
#     return query_idx, query, k_confident_idx, k_confident_query
#
#
# def de_ue_ner(
#     classifiers,
#     X_pool: Union[np.ndarray, Dataset, TransformersDataset],
#     n_instances: int,
#     gamma_or_k_confident_to_save: float = 0.25,
#     T: float = None,
#     sampling_type: str = None,
#     select_by_number_of_tokens=True,
#     require_sampling: bool = True,
#     method="bald",
#     **kwargs,
# ):
#     prob_arr = np.stack(
#         [classifier.predict_proba(X_pool) for classifier in classifiers], axis=1
#     )
#     if method == "bald":
#         scores = calculate_bald_score_ner(prob_arr)
#     # elif method == "vr":
#     #     scores = var_ratio(prob_arr)
#     elif method == "mean_ent":
#         scores = mean_entropy_ner(prob_arr)
#     # elif method == "max_prob":
#     #     scores = sampled_max_prob(prob_arr)
#     elif method == "prob_var":
#         scores = probability_variance_ner(prob_arr)
#     else:
#         raise NotImplementedError
#
#     argsort = np.argsort(-scores)
#     # How we select instances from the pool: with `__getitem__` (for np.ndarray) or with `select` (for Dataset)
#     method_name = "__getitem__" if isinstance(X_pool, np.ndarray) else "select"
#     select_fn = getattr(X_pool, method_name)
#
#     if select_by_number_of_tokens:
#         query_idx = get_query_idx_for_selecting_by_number_of_tokens(
#             X_pool, argsort, n_instances
#         )
#     else:
#         query_idx = argsort[:n_instances]
#
#     query = select_fn(query_idx)
#
#     if (sampling_type is None) or (not require_sampling):
#         return query_idx, query
#
#     elif sampling_type == "naive" and select_by_number_of_tokens:
#         ### Get cumulative sum of tokens; Determine how many we need to query;
#         ### Since a part of them will be queried anyway, take as many tokens, as the query will contain;
#         ### Find idx of the last instance that need to be queried; Query (including the last one).
#         cumsum_tokens = np.cumsum([len(x["tokens"].tokens) for x in select_fn(argsort)])
#         num_tokens_to_select = gamma_or_k_confident_to_save * cumsum_tokens[-1]
#         tokens_bound = num_tokens_to_select + sum(
#             [len(x["tokens"].tokens) for x in select_fn(query_idx)]
#         )
#         idx_last_plus_one = np.argwhere(cumsum_tokens > tokens_bound)[0][0] + 1
#         k_confident_idx = argsort[len(query_idx) : idx_last_plus_one]
#     elif sampling_type == "naive":  # not select by tokens
#         k_confident_idx = argsort[
#             n_instances : int(gamma_or_k_confident_to_save * len(argsort)) + n_instances
#         ]
#     elif sampling_type == "ups":
#         k_confident_idx = sample_ups(
#             argsort, query_idx, gamma_or_k_confident_to_save, T
#         )
#     elif sampling_type == "random":
#         idx_to_select_from = np.setdiff1d(argsort, query_idx)
#         share_to_sample = gamma_or_k_confident_to_save
#         k_confident_idx = np.random.choice(
#             idx_to_select_from,
#             int(share_to_sample * len(idx_to_select_from)),
#             replace=False,
#         )
#     else:
#         raise NotImplementedError
#
#     k_confident_query = select_fn(k_confident_idx)
#     return query_idx, query, k_confident_idx, k_confident_query
#
#
# # TODO: refactor
# def bald_sampling(
#     model,
#     X_pool,
#     n_instances,
#     gamma_or_k_confident_to_save=0.25,
#     T=None,
#     sampling_type=None,
#     select_by_number_of_tokens=True,
#     require_sampling=True,
#     only_head_dropout=True,
#     mc_iterations=10,
#     **kwargs,
# ):
#     if only_head_dropout:
#         prob_arr = np.array(
#             model.predict_proba(X_pool, mc_dropout=True, mc_iterations=mc_iterations)
#         )
#     else:
#         model.enable_dropout()
#         prob_arr = np.stack(
#             [model.predict_proba(X_pool) for _ in range(mc_iterations)], axis=0
#         )
#
#     scores = calculate_bald_score_ner(np.array(prob_arr))
#     argsort = np.argsort(-scores)
#
#     # How we select instances from the pool: with `__getitem__` (for np.ndarray) or with `select` (for Dataset)
#     method_name = "__getitem__" if isinstance(X_pool, np.ndarray) else "select"
#     select_fn = getattr(X_pool, method_name)
#
#     if select_by_number_of_tokens:
#         query_idx = get_query_idx_for_selecting_by_number_of_tokens(
#             X_pool, argsort, n_instances
#         )
#     else:
#         query_idx = argsort[:n_instances]
#
#     query = select_fn(query_idx)
#
#     if (sampling_type is None) or (not require_sampling):
#         return query_idx, query
#
#     elif sampling_type == "naive" and select_by_number_of_tokens:
#         ### Get cumulative sum of tokens; Determine how many we need to query;
#         ### Since a part of them will be queried anyway, take as many tokens, as the query will contain;
#         ### Find idx of the last instance that need to be queried; Query (including the last one).
#         cumsum_tokens = np.cumsum([len(x["tokens"].tokens) for x in select_fn(argsort)])
#         num_tokens_to_select = gamma_or_k_confident_to_save * cumsum_tokens[-1]
#         tokens_bound = num_tokens_to_select + sum(
#             [len(x["tokens"].tokens) for x in select_fn(query_idx)]
#         )
#         idx_last_plus_one = np.argwhere(cumsum_tokens > tokens_bound)[0][0] + 1
#         k_confident_idx = argsort[len(query_idx) : idx_last_plus_one]
#     elif sampling_type == "naive":  # not select by tokens
#         k_confident_idx = argsort[
#             n_instances : int(gamma_or_k_confident_to_save * len(argsort)) + n_instances
#         ]
#     elif sampling_type == "ups":
#         k_confident_idx = sample_ups(
#             argsort, query_idx, gamma_or_k_confident_to_save, T
#         )
#     elif sampling_type == "random":
#         idx_to_select_from = np.setdiff1d(argsort, query_idx)
#         share_to_sample = gamma_or_k_confident_to_save
#         k_confident_idx = np.random.choice(
#             idx_to_select_from,
#             int(share_to_sample * len(idx_to_select_from)),
#             replace=False,
#         )
#     else:
#         raise NotImplementedError
#
#     k_confident_query = select_fn(k_confident_idx)
#     return query_idx, query, k_confident_idx, k_confident_query


def ddu_sampling_cv(
    classifier,
    X_pool: Union[Dataset, TransformersDataset],
    n_instances: int,
    X_train: Union[Dataset, TransformersDataset],
    **ddu_kwargs,
):
    kwargs = dict(
        # Necessary
        model_wrapper=classifier,
        data_train=X_train,
        data_test=X_pool,
        # General
        data_is_tokenized=False,
        data_config=None,
        batch_size=classifier._batch_size_kwargs.eval_batch_size,
        # DDU
        use_activation=ddu_kwargs["use_activation"],
        use_spectralnorm=ddu_kwargs["use_spectralnorm"],
        to_numpy=True,
    )
    scores = calculate_ddu_scores_cv(**kwargs)
    uncertainty_estimates = 1 - scores
    argsort = np.argpartition(scores, kth=n_instances)
    query_idx = argsort[:n_instances]
    query = take_idx(X_pool, query_idx)

    return query_idx, query, uncertainty_estimates


def get_top_k_scorers(scores_N, batch_size, uncertainty=True):
    N = len(scores_N)
    batch_size = min(batch_size, N)
    candidate_scores, candidate_indices = torch.topk(
        scores_N, batch_size, largest=uncertainty
    )
    return candidate_scores.tolist(), candidate_indices.tolist()



import gc
import time
from typing import Union, Tuple, List
import os
import numpy as np
import logging
from datasets.metric import Metric

from torch import cuda

from .transformers_base_wrapper import TransformersBaseWrapper
from .wrapper_seq2seq import WrapperSeq2Seq
from ...utils.seq2seq_metrics import decode_all, calculate_bleu, calculate_additional_hf_metrics
from ...utils.summarization_metrics import (
    prepare_data_for_abssum_metrics,
    calculate_rouge,
    calculate_summac_score,
    calculate_bartscore,
    calculate_abstractiveness,
)


log = logging.getLogger()
USE_ADD_ATS_METRICS = os.environ.get("USE_ADD_ATS_METRICS", True)

if USE_ADD_ATS_METRICS:
    try:
        from ...utils.summarization_metrics import BARTScorer, SummaCZS

        ADD_METRICS_IMPORTED = True
    except Exception as e:
        log.info(f"Cannot import BARTScore & SummacZS. Exception:\n{e}")
        ADD_METRICS_IMPORTED = False


class WrapperAts(WrapperSeq2Seq, TransformersBaseWrapper):
    @staticmethod
    def get_tokenize_function(
        tokenizer, text_name, label_name, test_mode: bool = False, **kwargs
    ):
        if test_mode:

            def tokenize_function(instances):
                encoded = tokenizer(instances[text_name], truncation=True)
                return encoded

        else:

            def tokenize_function(instances):
                encoded = tokenizer(instances[text_name], truncation=True)
                if not test_mode:
                    with tokenizer.as_target_tokenizer():
                        labels = tokenizer(instances[label_name], truncation=True)
                    encoded["labels"] = labels["input_ids"]
                return encoded

        return tokenize_function

    @staticmethod
    def get_metric_name():
        return "test_rouge1"

    @staticmethod
    def get_datasets_metric_name():
        return "rouge"

    @staticmethod
    def get_additional_datasets_metric_names():
        return ["sacrebleu"]

    def get_kwargs_for_compute_metrics_fn(self):
        return {
            "tokenizer": self.tokenizer,
            "use_additional_metrics": USE_ADD_ATS_METRICS,
        }

    @staticmethod
    def get_compute_metrics_function(
        metric,
        tokenizer,
        additional_metrics: Union[Tuple[Metric], List[Metric]] = tuple(),
        use_additional_metrics: bool = True,
        **bartscorer_init_params,
    ):
        assert tokenizer is not None, "Please provide tokenizer"

        def compute_metrics(eval_preds):
            decoded_preds, decoded_labels, decoded_texts = decode_all(
                eval_preds, tokenizer, use_additional_metrics=use_additional_metrics
            )
            (
                predictions,
                labels,
                is_zeroword,
                is_uniword,
                tokenized,
            ) = prepare_data_for_abssum_metrics(
                decoded_preds, decoded_labels, decoded_texts
            )
            # ROUGE metrics
            result = calculate_rouge(predictions, labels, is_zeroword, is_uniword)
            # Add BLEU metric
            bleu_scores = calculate_bleu(decoded_preds, decoded_labels)
            result.update(bleu_scores)
            # Add mean generated length
            prediction_lens = [
                np.count_nonzero(pred != tokenizer.pad_token_id) for pred in predictions
            ]
            result["gen_len"] = np.mean(prediction_lens)

            if use_additional_metrics and decoded_texts is not None:
                abstr_scores = calculate_abstractiveness(
                    tokenized["preds"], tokenized["labels"], tokenized["texts"]
                )
                result.update(abstr_scores)
                if ADD_METRICS_IMPORTED:
                    # BARTScore
                    scorer = BARTScorer(**bartscorer_init_params)
                    # probable error with texts
                    bart_scores = calculate_bartscore(
                        decoded_preds, decoded_labels, decoded_texts, scorer
                    )
                    del scorer
                    cuda.empty_cache()
                    gc.collect()
                    # SummaC-ZC score
                    scorer = SummaCZS(granularity="sentence", model_name="vitc")
                    summac_scores = calculate_summac_score(
                        decoded_preds, decoded_labels, decoded_texts, scorer
                    )
                    result.update(bart_scores)
                    result.update(summac_scores)

            add_result = calculate_additional_hf_metrics(additional_metrics, decoded_labels, decoded_preds)
            result.update(add_result)

            return result

        return compute_metrics



import json
import logging
import os
from math import ceil
from pathlib import Path
from typing import List, Union

import dill
import numpy as np
import torch
import torch.multiprocessing as mp
from torch import optim
from tqdm import tqdm

log = logging.getLogger()


def point_to_device(point, device):
    if isinstance(point, torch.Tensor):
        return point.to(device)
    elif isinstance(point, list):
        return [el.to(device) for el in point]
    elif isinstance(point[next(iter(point.keys()))], list):
        return {k: torch.stack(v, dim=1).to(device) for k, v in point.items()}
    return {k: v.to(device) for k, v in point.items()}


def calculate_grad(model, optimizer, point):

    optimizer.zero_grad()

    point = point_to_device(point, next(model.parameters()).device)
    point = {k: v for k, v in point.items() if k in model.forward.__code__.co_varnames}
    loss = model(**point)["loss"]

    loss.backward()
    loss_gradient = torch.cat(
        [param.grad.view(-1) for param in model.parameters() if param.grad is not None]
    )
    return loss_gradient


def calculate_grad_flair(model, optimizer, point):
    optimizer.zero_grad()

    loss = model.forward_loss(point)
    loss = loss[0]

    loss.backward()
    loss_gradient = torch.cat(
        [param.grad.view(-1) for param in model.parameters() if param.grad is not None]
    )
    return loss_gradient


def calculate_tracin_score_one_epoch(model, optimizer, train_point, test_point, nu=1):

    train_grad = calculate_grad(model, optimizer, train_point)
    test_grad = calculate_grad(model, optimizer, test_point)
    return nu * (train_grad @ test_grad)


def load_dataloader(dataloader_path):
    with open(dataloader_path, "rb") as f:
        dataloader = dill.load(f)
    return dataloader


def calculate_tracin_outlier_score_one_model(
    model_load_func,
    args: Union,  # args for model_load function
    weights_path,
    dataloader_path,
    work_dir,
    n_epoch,
    nu=1,
    cuda_device=0,
    framework="transformers",
):

    log.info("Loading model and dataloader")
    model = model_load_func(*args)
    if framework == "flair":
        import flair

        # bilstm-crf case
        # set device before creating model
        flair.device = torch.device(f"cuda:{cuda_device}")
        model.load(weights_path)
        model.to(torch.device(f"cuda:{cuda_device}"))
        # also manually move all embeddings to device
        for embedding in model.embeddings.embeddings:
            embedding.to(torch.device(f"cuda:{cuda_device}"))
    else:
        model.load_state_dict(torch.load(weights_path))
        model.cuda(cuda_device)

    dataloader = load_dataloader(dataloader_path)
    log.info("Done with loading model and dataloader")

    optimizer = optim.SGD(model.parameters(), lr=1)
    scores = []

    grad_func = calculate_grad_flair if framework == "flair" else calculate_grad
    log.info("Start calculating TracIn scores")
    for i, point in enumerate(tqdm(dataloader)):
        point_to_device(point, 0)

        point_grad = grad_func(model, optimizer, point)
        point_epoch_score = nu * (point_grad @ point_grad.T)

        scores.append(float(point_epoch_score.item()))
    log.info("Done with calculating TracIn scores")

    with open(Path(work_dir) / f"scores_epoch_{n_epoch}.json", "w") as f:
        json.dump(scores, f)
    return scores


def load_model(model_path: str or Path) -> torch.nn.Module:
    with open(model_path, "rb") as f:
        model = dill.load(f)
    return model


def load_flair_model(model_path, idx_to_tag, config):
    # here we create model, because we can't pickle it
    from ..al4nlp.models.bilstm_crf import create_flair_model_tokenizer

    cache_dir = Path(config.cache_dir) if config.cache_dir is not None else None
    model, tokenizer = create_flair_model_tokenizer(
        config.target_model,
        idx_to_tag,
        config.data.label_name,
        config.seed,
        cache_dir=cache_dir,
    )
    return model


def calculate_outlier_scores(
    model_path: str or Path,
    weights_paths: List[str or Path],
    dataloader_path: str or Path,
    work_dir: str or Path,
    max_num_processes: int = None,
    task: str = "ner",
    nu: float or int = 1,
    framework: str = "transformers",
    idx_to_tag: dict = None,
    config: dict = None,
):
    torch.cuda.empty_cache()
    mp.set_start_method("spawn", force=True)

    processes = []
    cuda_devices = list(map(int, os.environ["CUDA_VISIBLE_DEVICES"].split(",")))
    total_procs = len(weights_paths)
    num_cudas = len(cuda_devices)
    num_procs_per_cuda = min(max_num_processes, ceil(len(weights_paths) / num_cudas))
    num_procs_per_core = ceil(
        total_procs / (max_num_processes * num_cudas)
    )  # ~ num batches

    load_model_func = load_flair_model if framework == "flair" else load_model
    model_args = (
        [model_path, idx_to_tag, config] if framework == "flair" else [model_path]
    )
    for n_proc_per_core in range(num_procs_per_core):
        for n_proc_per_cuda in range(num_procs_per_cuda):
            for i_cuda, cuda_device in enumerate(cuda_devices):
                n_epoch = (
                    i_cuda
                    + n_proc_per_cuda * num_cudas
                    + n_proc_per_core * (num_cudas * num_procs_per_cuda)
                )
                if n_epoch >= total_procs:
                    break
                log.info(
                    f"Starting process for epoch {n_epoch} on cuda device {cuda_device}, a process number "
                    f"{n_proc_per_cuda} on this cuda device."
                )
                p = mp.Process(
                    target=calculate_tracin_outlier_score_one_model,
                    args=(
                        load_model_func,
                        model_args,
                        weights_paths[n_epoch],
                        dataloader_path,
                        work_dir,
                        n_epoch,
                        nu,
                        i_cuda,
                        framework,
                    ),
                )
                p.start()
                processes.append(p)
                torch.cuda.empty_cache()
        for p in processes:
            p.join()

    log.info(f"TracIn scores calculation done.")
    scores = []
    for n_epoch in range(total_procs):
        with open(Path(work_dir) / f"scores_epoch_{n_epoch}.json") as f:
            scores.append(json.load(f))

    if task == "ner":
        dataloader = load_dataloader(dataloader_path)
        if framework == "flair":
            lengths = [len(point[0].tokens) for point in dataloader]
        else:
            lengths = [len(point["idx_first_bpe"][0]) for point in dataloader]
        final_scores = np.mean(scores, axis=0) / lengths
        return final_scores, scores, lengths

    elif task == "cls":
        final_scores = np.mean(scores, axis=0)
        return final_scores, scores
    else:
        raise NotImplementedError



from typing import Union

import numpy as np
import plotly.express as px
from scipy.stats import t

PRETRAINED_NAMES_MAPPING = {
    "distilbert": "DistilBERT",
    "bert-base": "BERT base",
    "bert-large": "BERT large",
    "google/electra-base-discriminator": "ELECTRA base",
    "roberta-base": "RoBERTa base",
    "roberta-large": "RoBERTa large",
    "xlnet-base": "XLNet base",
    "xlnet-large": "XLNet large",
    "distilrobert": "DistilRoBERTa",
    "microsoft/": "DeBERTa",
    "facebook/bart-base": "BART base",
    "facebook/bart-large": "BART large",
}

AL_EXPERIMENT_TYPES_MAPPING = {
    "one": {"cls": "LC", "ner": "MNLP", "ats": "NSP", "nmt": "NSP"},
    "asm": "A-S M",
    "astm": "PLASM",
    "plasm": "PLASM",
    "full": "Full data",
}

DEFAULT_METRIC_NAME = {
    "cls": "Accuracy",
    "ner": "F1",
    "ats": "ROUGE-L",
    "nmt": "BLEU",
}

COLORS = [
    "rgb(50, 150, 200)",
    "rgb(30, 180, 30)",
    "rgb(128, 0, 128)",
    "rgb(255, 165, 0)",
    "rgb(200, 128, 150)",
    "rgb(128, 128, 128)",
    "rgb(20, 200, 250)",
    "rgb(255, 15, 255)",
    "rgb(60, 180, 120)",
    "rgb(128, 0, 0)",
    "rgb(67, 32, 125)",
    "rgb(34, 150, 160)",
    "rgb(200, 200, 150)",
]


def _make_sleeve_color(color):
    return "rgba" + color[3:-1] + ", 0.2)"


def _sub_name(name):
    for pretrained_name, mapping in PRETRAINED_NAMES_MAPPING.items():
        if name.startswith(pretrained_name):
            return mapping
    return name


def _sub_al_experiment_type(name, task="cls"):
    for al_exp_type, mapping in AL_EXPERIMENT_TYPES_MAPPING.items():
        if name == al_exp_type:
            if isinstance(mapping, dict):
                return mapping[task]
            return mapping
    return name.title()


def annotate_line(columns_to_group_by, columns_to_substitute):
    """
    Function to generate automatic annotation for each line.
    `transformations_list` will consist of triples
    <initial_text, function_to_transform_value_in_this_column, idx_of_this_column>
    """
    transformations_list = []
    if "id_experiment" in columns_to_group_by:
        idx = columns_to_group_by.index("id_experiment")
        transformations_list.append(["exp. ", None, idx])
    if "metric_name" in columns_to_group_by:
        idx = columns_to_group_by.index("metric_name")
        transformations_list.append(["metric: ", None, idx])
    if "al_type" in columns_to_group_by:
        idx = columns_to_group_by.index("al_type")
        transformations_list.append(["", _sub_al_experiment_type, idx])
    if "strategy" in columns_to_group_by:
        idx = columns_to_group_by.index("strategy")
        transformations_list.append(["strat. ", None, idx])
    if "acquisition" in columns_to_group_by:
        idx = columns_to_group_by.index("acquisition")
        transformations_list.append(["acq. ", _sub_name, idx])
    if "target" in columns_to_group_by:
        idx = columns_to_group_by.index("target")
        transformations_list.append(["succ. ", _sub_name, idx])
    if "ups" in columns_to_group_by:
        idx = columns_to_group_by.index("ups")
        transformations_list.append(["", None, idx])
    if "deleted" in columns_to_group_by:
        idx = columns_to_group_by.index("deleted")
        transformations_list.append(["del. ", str, idx])

    new_columns = []
    for multicolumn in columns_to_substitute:
        if isinstance(multicolumn, str):
            multicolumn = [multicolumn]
        result = ""
        for (pre_text, func, idx) in transformations_list:
            text = func(multicolumn[idx]) if func is not None else multicolumn[idx]
            result += pre_text + str(text) + ", "
        result = result[:-2]  # Remove the last comma and space
        new_columns.append(result)
    return new_columns


def plot_with_confidence_interval(
    df,
    columns_to_group_by=(
        "al_type",
        "strategy",
        "acquisition",
        "successor",
        "target",
    ),  # columns to groupby
    save_name="",  # how to save the img. if empty, no save is done
    metric_name=None,  # name for the metric on the y-scale
    task="cls",  # "cls" / "ner" / "ats" / "nmt"
    sleeve_type="conf",  # how to plot the variance sleeve. "conf" for conf interval / None for no conf / any for ±std
    title="",  # plot title
    columns=None,  # how to name columns (otherwise names automatically)
    index=None,  # must have length equal to num_queries
    colors=None,  # colors for lines and sleeves
    line_dash_sequence=None,  # how to plot different lines
    legend_font_size=20,  # size of the legend
    labels_font_size=15,  # size of the x_label / y_label text
    x_label: str = None,  # name of x axis
    y_label: str = None,  # name of y axis
    remove_whitespace: bool = True,  # whether to remove whitespaces around the figure when saving it
    x_title_standoff: Union[
        int, float
    ] = None,  # if not None, the margin between x-label and the axis
    y_title_standoff: Union[
        int, float
    ] = None,  # if not None, the margin between y-label and the axis
    **kwargs,  # other kwargs for px.line
):
    """
    sleeve_type == "conf" or "std" or None. if "conf", `count` should be specified
    """
    plot_sleeve = sleeve_type not in ["none", None, "no"]
    if metric_name is None:
        metric_name = DEFAULT_METRIC_NAME[task]
    metric_name = f"Performance, {metric_name.title()}"
    # Remove potentially redundant columns
    for column in ["id_experiment", "deleted", "seed"]:
        if (column in df) and (column not in columns_to_group_by):
            df = df.drop(columns=column)
    # Group by the specified columns
    groupby = df.groupby(list(columns_to_group_by))
    df_mean = groupby.mean().T
    df_std = groupby.std().T

    if sleeve_type == "conf":
        df_count = groupby.count().T
        columns_to_save = [x for x in df_count.index if x.startswith("iter")]
        df_count = df_count.loc[columns_to_save]
        # Calculate components for conf interval
        conf_lev_vals = t.ppf(0.975, df_count)
        std_by_conf_val = df_std * conf_lev_vals
        counts_sqrt = df_count ** 0.5
        # Calculate lower and upper bounds
        df_lb = df_mean - std_by_conf_val / counts_sqrt
        df_ub = df_mean + std_by_conf_val / counts_sqrt
    elif not plot_sleeve:
        df_lb = df_mean - df_std
        df_ub = df_mean + df_std
    else:
        df_ub = df_lb = df_mean

    # Name columns if they are not specified
    if columns is None:
        columns = annotate_line(columns_to_group_by, df_mean.columns)
    df_mean.columns = df_ub.columns = df_lb.columns = columns

    # Get percent of labeled data so far if not specified
    if index is None:
        index = np.arange(1, len(df_mean) + 1)
        if task == "ner":
            index *= 2
        elif task in ["ats", "nmt"]:
            index *= 10
    df_mean.index = df_lb.index = df_ub.index = index
    # Label axes
    if x_label is None:
        if task in ["ats", "nmt"]:
            x_label = "Num. labeled instances"
        else:
            x_label = "Labeled Data, %"
    if y_label is None:
        y_label = metric_name

    if line_dash_sequence is None:
        line_dash_sequence = ["dash", "dashdot", "longdashdot", "solid", "longdash"]

    # Reset index, melt and rename the column
    df_to_plot = df_mean.reset_index()
    df_to_plot = df_to_plot.melt(id_vars=["index"], value_name=metric_name)
    df_to_plot = df_to_plot.rename(columns={"index": x_label, metric_name: y_label})

    # Add sleeve colors if neither they nor colors are specified
    if colors is None:
        colors = COLORS
    sleeve_colors = list(map(_make_sleeve_color, colors))

    # Plot the lines
    fig = px.line(
        df_to_plot,
        x=x_label,
        y=y_label,
        line_dash="variable",
        line_dash_sequence=line_dash_sequence,
        title=title,
        color="variable",
        color_discrete_sequence=colors,
        **kwargs,
    )

    if plot_sleeve:
        # Plot sleeves
        for i, name in enumerate(df_mean.columns):
            # Upper bound
            fig.add_scatter(
                x=df_ub.index,
                y=df_ub.iloc[:, i],
                name=name + "_upper_bound",
                marker=dict(color="#444", size=0),
                line=dict(width=0),
                showlegend=False,
                mode="lines",
            )
            # Lower bound
            fig.add_scatter(
                x=df_lb.index,
                y=df_lb.iloc[:, i],
                name=name + "_lower_bound",
                marker=dict(color="#444"),
                line=dict(width=0),
                fillcolor=sleeve_colors[i],
                fill="tonexty",
                showlegend=False,
                mode="lines",
            )

    # Update legend position & font
    labels_font_size = 15 if labels_font_size is None else labels_font_size
    fig.update_layout(
        legend=dict(
            yanchor="bottom",
            y=0.01,
            xanchor="right",
            x=0.99,
            font=dict(size=legend_font_size),
        ),
        font=dict(size=labels_font_size),
        legend_title_text=None,
    )
    # Title standoff
    if x_title_standoff is not None:
        fig.update_xaxes(title_standoff=x_title_standoff)
    if y_title_standoff is not None:
        fig.update_yaxes(title_standoff=y_title_standoff)

    # Save
    if save_name != "" and save_name is not None:
        if remove_whitespace:
            fig.update_layout(margin=dict(l=0, r=0, t=10, b=1),)
        fig.write_image(save_name)

    return fig



from .gmm import compute_density, get_gmm_log_probs, gmm_fit
from .mahalanobis import (
    compute_centroids,
    compute_inv_covariance,
    mahalanobis_distance_with_known_centroids_sigma_inv,
)
from .mahalanobis_v2 import (
    compute_inv_covariance_v2,
    mahalanobis_distance_with_known_centroids_sigma_inv_v2,
)
from .batchbald.batchbald import get_batchbald_batch
from .batchbald.consistent_dropout import make_dropouts_consistent
from .ue import UeEstimatorHybrid, TextClassifier



import logging
import time

import numpy as np
from tqdm import tqdm

log = logging.getLogger()


def compute_centroids(train_features, train_labels, class_cond=True):
    if class_cond:
        centroids = []
        for label in np.sort(np.unique(train_labels)):
            centroids.append(train_features[train_labels == label].mean(axis=0))
        return np.asarray(centroids)
    else:
        return train_features.mean(axis=0)


def compute_covariance(centroids, train_features, train_labels, class_cond=True):
    cov = np.zeros((train_features.shape[1], train_features.shape[1]))
    if class_cond:
        for c, mu_c in tqdm(enumerate(centroids)):
            for x in train_features[train_labels == c]:
                d = (x - mu_c)[:, None]
                cov += d @ d.T
    else:
        for x in train_features:
            d = (x - centroids[0])[:, None]
            cov += d @ d.T
    cov /= train_features.shape[0]

    try:
        sigma_inv = np.linalg.inv(cov)
    except:
        sigma_inv = np.linalg.pinv(cov)
        log.info("Compute pseudo-inverse matrix")

    return sigma_inv


def mahalanobis_distance(
    train_features,
    train_labels,
    eval_features,
    centroids=None,
    covariance=None,
    return_full=False,
):
    if centroids is None:
        centroids = compute_centroids(train_features, train_labels)
    if covariance is None:
        covariance = compute_covariance(centroids, train_features, train_labels)
    diff = eval_features[:, None, :] - centroids[None, :, :]
    start = time.time()
    dists = np.matmul(np.matmul(diff, covariance), diff.transpose(0, 2, 1))
    dists = np.asarray([np.diag(dist) for dist in dists])
    end = time.time()
    if return_full:
        return dists, end - start
    else:
        return np.min(dists, axis=1), end - start


def mahalanobis_distance_marginal(
    train_features, train_labels, eval_features, centroids=None, covariance=None
):
    if centroids is None:
        centroids = compute_centroids(train_features, train_labels, class_cond=False)
    if covariance is None:
        covariance = compute_covariance(
            centroids, train_features, train_labels, class_cond=False
        )

    diff = eval_features - centroids[0][None, :]
    dists = np.matmul(np.matmul(diff, covariance), diff.T)
    return np.diag(dists)


def mahalanobis_distance_relative(
    train_features,
    train_labels,
    eval_features,
    centroids=None,
    covariance=None,
    train_centroid=None,
    train_covariance=None,
):
    if centroids is None:
        centroids = compute_centroids(train_features, train_labels)
    if covariance is None:
        covariance = compute_covariance(centroids, train_features, train_labels)
    diff = eval_features[:, None, :] - centroids[None, :, :]
    dists = np.matmul(np.matmul(diff, covariance), diff.transpose(0, 2, 1))
    dists = np.asarray([np.diag(dist) for dist in dists])

    md_marginal = mahalanobis_distance_marginal(
        train_features, eval_features, train_centroid, train_covariance
    )
    return np.min(dists - md_marginal[:, None], axis=1)



import torch
from toma import toma
from tqdm.auto import tqdm

from .utils import gather_expand, batch_multi_choices


class JointEntropy:
    """Random variables (all with the same # of categories $C$) can be added via `JointEntropy.add_variables`.

    `JointEntropy.compute` computes the joint entropy.

    `JointEntropy.compute_batch` computes the joint entropy of the added variables with each of the variables in the provided batch probabilities in turn."""

    def compute(self) -> torch.Tensor:
        """Computes the entropy of this joint entropy."""
        raise NotImplementedError()

    def add_variables(self, log_probs_N_K_C: torch.Tensor) -> "JointEntropy":
        """Expands the joint entropy to include more terms."""
        raise NotImplementedError()

    def compute_batch(
        self, log_probs_B_K_C: torch.Tensor, output_entropies_B=None
    ) -> torch.Tensor:
        """Computes the joint entropy of the added variables together with the batch (one by one)."""
        raise NotImplementedError()


class ExactJointEntropy(JointEntropy):
    joint_probs_M_K: torch.Tensor

    def __init__(self, joint_probs_M_K: torch.Tensor):
        self.joint_probs_M_K = joint_probs_M_K

    @staticmethod
    def empty(K: int, device=None, dtype=None) -> "ExactJointEntropy":
        return ExactJointEntropy(torch.ones((1, K), device=device, dtype=dtype))

    def compute(self) -> torch.Tensor:
        probs_M = torch.mean(self.joint_probs_M_K, dim=1, keepdim=False)
        nats_M = -torch.log(probs_M) * probs_M
        entropy = torch.sum(nats_M)
        return entropy

    def add_variables(self, log_probs_N_K_C: torch.Tensor) -> "ExactJointEntropy":
        assert self.joint_probs_M_K.shape[1] == log_probs_N_K_C.shape[1]

        N, K, C = log_probs_N_K_C.shape
        joint_probs_K_M_1 = self.joint_probs_M_K.t()[:, :, None]

        probs_N_K_C = log_probs_N_K_C.exp()

        # Using lots of memory.
        for i in range(N):
            probs_i__K_1_C = probs_N_K_C[i][:, None, :].to(
                joint_probs_K_M_1, non_blocking=True
            )
            joint_probs_K_M_C = joint_probs_K_M_1 * probs_i__K_1_C
            joint_probs_K_M_1 = joint_probs_K_M_C.reshape((K, -1, 1))

        self.joint_probs_M_K = joint_probs_K_M_1.squeeze(2).t()
        return self

    def compute_batch(self, log_probs_B_K_C: torch.Tensor, output_entropies_B=None):
        assert self.joint_probs_M_K.shape[1] == log_probs_B_K_C.shape[1]

        B, K, C = log_probs_B_K_C.shape
        M = self.joint_probs_M_K.shape[0]

        if output_entropies_B is None:
            output_entropies_B = torch.empty(
                B, dtype=log_probs_B_K_C.dtype, device=log_probs_B_K_C.device
            )

        pbar = tqdm(total=B, desc="ExactJointEntropy.compute_batch", leave=False)

        @toma.execute.chunked(log_probs_B_K_C, initial_step=1024, dimension=0)
        def chunked_joint_entropy(
            chunked_log_probs_b_K_C: torch.Tensor, start: int, end: int
        ):
            chunked_probs_b_K_C = chunked_log_probs_b_K_C.exp()
            b = chunked_probs_b_K_C.shape[0]

            probs_b_M_C = torch.empty(
                (b, M, C),
                dtype=self.joint_probs_M_K.dtype,
                device=self.joint_probs_M_K.device,
            )
            for i in range(b):
                torch.matmul(
                    self.joint_probs_M_K,
                    chunked_probs_b_K_C[i].to(self.joint_probs_M_K, non_blocking=True),
                    out=probs_b_M_C[i],
                )
            probs_b_M_C /= K

            output_entropies_B[start:end].copy_(
                torch.sum(-torch.log(probs_b_M_C) * probs_b_M_C, dim=(1, 2)),
                non_blocking=True,
            )

            pbar.update(end - start)

        pbar.close()

        return output_entropies_B


class SampledJointEntropy(JointEntropy):
    """Random variables (all with the same # of categories $C$) can be added via `SampledJointEntropy.add_variables`.

    `SampledJointEntropy.compute` computes the joint entropy.

    `SampledJointEntropy.compute_batch` computes the joint entropy of the added variables with each of the variables in the provided batch probabilities in turn."""

    sampled_joint_probs_M_K: torch.Tensor

    def __init__(
        self, sampled_joint_probs_M_K: torch.Tensor, scaling_factor: torch.Tensor = None
    ):
        if scaling_factor is not None:
            self.scaling_factor = scaling_factor
            self.sampled_joint_probs_M_K = sampled_joint_probs_M_K
        else:
            self.scaling_factor = self._calculate_scaling_factor(
                sampled_joint_probs_M_K
            )
            self.sampled_joint_probs_M_K = sampled_joint_probs_M_K / self.scaling_factor

    @staticmethod
    def _calculate_scaling_factor(
        sampled_joint_probs_M_K: torch.Tensor,
    ) -> torch.Tensor:
        scaling_factor = 1.0
        scaled_probs = sampled_joint_probs_M_K.clone()
        while scaled_probs.max() < 1e-2:
            scaling_factor /= 10
            scaled_probs *= 10
        return torch.Tensor([scaling_factor]).to(sampled_joint_probs_M_K)

    @staticmethod
    def empty(K: int, device=None, dtype=None) -> "SampledJointEntropy":
        return SampledJointEntropy(torch.ones((1, K), device=device, dtype=dtype))

    @staticmethod
    def sample(probs_N_K_C: torch.Tensor, M: int) -> "SampledJointEntropy":
        K = probs_N_K_C.shape[1]

        # S: num of samples per w
        S = M // K

        # Samples depending on each "sample" of one observation:
        # e.g. have 2 samples of 1 obs with probas [[0.6, 0.3, 0.1], [0.1, 0.8, 0.1]]
        # Most probable sample output will be [0, 1]. Consequently, the first column of
        # `samples_M_K` will depend on [0.6, 0.3], while the second - on [0.1, 0.8]
        choices_N_K_S = batch_multi_choices(probs_N_K_C, S).long()

        expanded_choices_N_1_K_S = choices_N_K_S[:, None, :, :]
        expanded_probs_N_K_1_C = probs_N_K_C[:, :, None, :]

        probs_N_K_K_S = gather_expand(
            expanded_probs_N_K_1_C, dim=-1, index=expanded_choices_N_1_K_S
        )
        # exp sum log seems necessary to avoid 0s?
        log_sum_K_K_S = torch.sum(torch.log(probs_N_K_K_S), dim=0, keepdim=False)
        scaling_factor = log_sum_K_K_S.max()
        probs_K_K_S = torch.exp(log_sum_K_K_S - scaling_factor)
        samples_K_M = probs_K_K_S.reshape((K, -1))

        samples_M_K = samples_K_M.t()
        return SampledJointEntropy(samples_M_K, scaling_factor.exp())

    def compute(self) -> torch.Tensor:
        sampled_joint_probs_M = torch.mean(
            self.sampled_joint_probs_M_K, dim=1, keepdim=False
        )
        nats_M = -(torch.log(sampled_joint_probs_M) + self.scaling_factor.log())
        entropy = torch.mean(nats_M)
        return entropy

    def add_variables(
        self, log_probs_N_K_C: torch.Tensor, M2: int
    ) -> "SampledJointEntropy":
        K = log_probs_N_K_C.shape[1]
        assert self.sampled_joint_probs_M_K.shape[1] == K

        sample_K_M1_1 = self.sampled_joint_probs_M_K.t()[:, :, None]

        new_sample_M2_K = self.sample(log_probs_N_K_C.exp(), M2).sampled_joint_probs_M_K
        new_sample_K_1_M2 = new_sample_M2_K.t()[:, None, :]

        merged_sample_K_M1_M2 = sample_K_M1_1 * new_sample_K_1_M2
        merged_sample_K_M = merged_sample_K_M1_M2.reshape((K, -1))

        self.sampled_joint_probs_M_K = merged_sample_K_M.t()

        return self

    def compute_batch(self, log_probs_B_K_C: torch.Tensor, output_entropies_B=None):
        assert self.sampled_joint_probs_M_K.shape[1] == log_probs_B_K_C.shape[1]

        B, K, C = log_probs_B_K_C.shape
        M = self.sampled_joint_probs_M_K.shape[0]

        if output_entropies_B is None:
            output_entropies_B = torch.empty(
                B, dtype=log_probs_B_K_C.dtype, device=log_probs_B_K_C.device
            )

        pbar = tqdm(total=B, desc="SampledJointEntropy.compute_batch", leave=False)

        @toma.execute.chunked(log_probs_B_K_C, initial_step=1024, dimension=0)
        def chunked_joint_entropy(
            chunked_log_probs_b_K_C: torch.Tensor, start: int, end: int
        ):
            b = chunked_log_probs_b_K_C.shape[0]
            chunked_probs_b_K_C = chunked_log_probs_b_K_C.exp()

            probs_b_M_C = torch.empty(
                (b, M, C),
                dtype=self.sampled_joint_probs_M_K.dtype,
                device=self.sampled_joint_probs_M_K.device,
            )
            torch.matmul(
                self.sampled_joint_probs_M_K,
                chunked_probs_b_K_C.to(self.sampled_joint_probs_M_K, non_blocking=True),
                out=probs_b_M_C,
            )
            probs_b_M_C /= K

            q_1_M_1 = self.sampled_joint_probs_M_K.mean(dim=1, keepdim=True)[None]

            output_entropies_B[start:end].copy_(
                torch.sum(
                    -(torch.log(probs_b_M_C) + self.scaling_factor.log())
                    * probs_b_M_C
                    / q_1_M_1,
                    dim=(1, 2),
                )
                / M,
                non_blocking=True,
            )

            pbar.update(end - start)

        pbar.close()

        return output_entropies_B


class DynamicJointEntropy(JointEntropy):
    inner: JointEntropy
    log_probs_max_N_K_C: torch.Tensor
    N: int
    M: int

    def __init__(self, M: int, max_N: int, K: int, C: int, dtype=None, device=None):
        self.M = M
        self.N = 0
        self.max_N = max_N

        self.inner = ExactJointEntropy.empty(K, dtype=dtype, device=device)
        self.log_probs_max_N_K_C = torch.empty(
            (max_N, K, C), dtype=dtype, device=device
        )

    def add_variables(self, log_probs_N_K_C: torch.Tensor) -> "DynamicJointEntropy":
        C = self.log_probs_max_N_K_C.shape[2]
        add_N = log_probs_N_K_C.shape[0]

        assert self.log_probs_max_N_K_C.shape[0] >= self.N + add_N
        assert self.log_probs_max_N_K_C.shape[2] == C

        self.log_probs_max_N_K_C[self.N : self.N + add_N] = log_probs_N_K_C
        self.N += add_N

        num_exact_samples = C ** self.N
        if num_exact_samples > self.M:
            self.inner = SampledJointEntropy.sample(
                self.log_probs_max_N_K_C[: self.N].exp(), self.M
            )
        else:
            self.inner.add_variables(log_probs_N_K_C)

        return self

    def compute(self) -> torch.Tensor:
        return self.inner.compute()

    def compute_batch(
        self, log_probs_B_K_C: torch.Tensor, output_entropies_B=None
    ) -> torch.Tensor:
        """Computes the joint entropy of the added variables together with the batch (one by one)."""
        return self.inner.compute_batch(log_probs_B_K_C, output_entropies_B)



from datasets import concatenate_datasets, Dataset
import numpy as np
import logging
from pathlib import Path
import os
import json
from omegaconf import DictConfig
from typing import Union

from acleto.al4nlp.utils.transformers_dataset import TransformersDataset


log = logging.getLogger()


def restore_queries(
        from_checkpoint: Union[DictConfig, str, Path],
        train_instances: Union[Dataset, TransformersDataset],
        initial_data: Union[Dataset, TransformersDataset] = None,
        unlabeled_data: Union[Dataset, TransformersDataset] = None,
        text_column: str = "text"
):
    if isinstance(from_checkpoint, (Path, str)):
        work_dir = Path(from_checkpoint)
        last_iter = None
    else:
        work_dir = Path(from_checkpoint["path"])
        last_iter = from_checkpoint.get("last_iteration", None)

    ids_paths = sorted(
        [x for x in os.listdir(work_dir) if x.startswith("ids_data_query_")]
    )
    if last_iter is not None:
        ids_paths = ids_paths[: last_iter + 1]
    id_first_iteration = len(ids_paths) - 1
    log.info(
        f"Resuming active learning from iteration {id_first_iteration + 1}..."
    )

    for i in range(len(ids_paths)):
        ids_data_query_path = f"ids_data_query_{i}.json"
        with open(work_dir / ids_data_query_path) as f:
            query_ids = json.load(f)
        if ids_data_query_path == "ids_data_query_0.json":
            query = train_instances.select(query_ids)
            if initial_data is not None:
                assert (
                    query[text_column] == initial_data[text_column]
                ), "Initial iteration results differ!"
            else:
                initial_data = query
                unlabeled_data = train_instances.select(
                    np.setdiff1d(range(len(train_instances)), query_ids)
                )
        else:
            query = unlabeled_data.select(query_ids)
            if isinstance(initial_data, TransformersDataset):
                initial_data.add(query)
            elif isinstance(initial_data, Dataset):
                initial_data = concatenate_datasets([initial_data, query])
            else:
                initial_data = query
            unlabeled_data = unlabeled_data.select(
                np.setdiff1d(range(len(unlabeled_data)), query_ids)
            )
    log.info(
        f"Query ids up to iteration {id_first_iteration} inclusive successfully loaded."
    )
    return initial_data, unlabeled_data



import numpy as np
from typing import List, Tuple, Dict, Union
import time
from datasets import load_metric
import os
from pathlib import Path
import sys
import logging

from rouge_score import tokenize
from nltk.stem import porter
from nltk import ngrams
from nltk.tokenize import sent_tokenize

log = logging.getLogger()

if os.environ.get("USE_ADD_ATS_METRICS", True):
    try:
        path = Path("/".join(os.path.abspath(__file__).split("/")[:-1]))
        sys.path.append(str(path / "../../utils/packages/summac"))
        sys.path.append(str(path / "../../utils/packages/BARTScore"))
        from .packages.bart_score import BARTScorer
        from .packages.summac.summac.model_summac import SummaCZS

        log.info("Successfully loaded BARTScore and SummaC models.")

        ADD_METRICS_IMPORTED = True
    except:
        log.info(
            "Cannot load BARTScore & SummaC models;"
            " Consistency scores will not be computed for abstractive summarization."
        )
        ADD_METRICS_IMPORTED = False


def calculate_ngram_overlap(summary, text, n=1, use_modified=True):
    summary_ngrams = list(ngrams(summary, n))
    text_ngrams = list(ngrams(text, n))

    if len(summary_ngrams) > 0:
        ngrams_intersection = set(summary_ngrams).intersection(set(text_ngrams))
        if use_modified:
            word_is_part_of_ngram_copied = [
                any((x in ngram for ngram in ngrams_intersection)) for x in summary
            ]
            return 1 - sum(word_is_part_of_ngram_copied) / len(
                word_is_part_of_ngram_copied
            )
        else:
            return sum([x not in ngrams_intersection for x in summary_ngrams]) / len(
                summary_ngrams
            )
    return None


def calculate_rouge(
    predictions: List[str],
    labels: List[str],
    is_zeroword: np.ndarray = None,
    is_uniword: np.ndarray = None,
) -> Dict[str, np.ndarray]:
    """
    Calculate ROUGE scores
    :param predictions:
    :param labels:
    :param is_zeroword:
    :param is_uniword:
    :return: dictionary with 4 values
    """
    if is_zeroword is None:
        is_zeroword = np.zeros(len(predictions), dtype=bool)
    if is_uniword is None:
        is_uniword = np.zeros(len(predictions), dtype=bool)

    rouge = load_metric("rouge")
    rouges = rouge.compute(
        predictions=predictions,
        references=labels,
        use_stemmer=True,
        use_aggregator=False,
    )
    metrics = np.array([[x.fmeasure for x in value] for value in rouges.values()])
    # Substitute invalid observations with nans
    rouge1 = np.mean(metrics[0][~is_zeroword])
    rouge2 = np.mean(metrics[1][~(is_zeroword | is_uniword)])
    rougeL = np.mean(metrics[2][~is_zeroword])
    rougeLsum = np.mean(metrics[3][~is_zeroword])
    metrics = {
        "rouge1": rouge1,
        "rouge2": rouge2,
        "rougeL": rougeL,
        "rougeLsum": rougeLsum,
    }
    return metrics


def calculate_summac_score(
    predictions: List[str], labels: List[str], texts: List[str], scorer=None
) -> Dict[str, np.ndarray]:
    start_time = time.time()
    if scorer is None:
        scorer = SummaCZS(granularity="sentence", model_name="vitc")
    preds_score = np.mean(scorer.score(texts, predictions)["scores"])
    labels_score = np.mean(scorer.score(texts, labels)["scores"])
    log.info(f"SummaC took {time.time() - start_time:.4f} seconds")
    return {"SUMMAC-pred": preds_score, "SUMMAC-label": labels_score}


def calculate_bartscore(
    predictions, labels, texts, scorer=None, batch_size=4, **bartscorer_init_params
) -> Dict[str, np.ndarray]:

    # decoded_texts = ["\n".join(sent_tokenize(text.strip())) for text in texts]
    start_time = time.time()
    if scorer is None:
        scorer = BARTScorer(**bartscorer_init_params)
    scores = get_bart_scores(
        scorer,
        preds=predictions,
        refs=labels,
        texts=texts,
        batch_size=batch_size,
        aggregate=True,
    )
    log.info(f"BARTScore took {time.time() - start_time:.4f} seconds")
    return scores


def get_bart_scores(scorer, preds, refs, texts, batch_size=4, aggregate=True):
    scores = {}
    scores["BARTScore-sh"] = np.array(scorer.score(texts, preds, batch_size=batch_size))
    scores["BARTScore-rh"] = np.array(scorer.score(refs, preds, batch_size=batch_size))
    scores["BARTScore-hr"] = np.array(scorer.score(preds, refs, batch_size=batch_size))
    scores["BARTScore-f1"] = (
        2
        * scores["BARTScore-rh"]
        * scores["BARTScore-hr"]
        / (scores["BARTScore-rh"] + scores["BARTScore-hr"])
    )
    scores["BARTScore-fa"] = (scores["BARTScore-rh"] + scores["BARTScore-hr"]) / 2

    if aggregate:
        scores = {key: np.mean(value) for key, value in scores.items()}
    return scores


def prepare_data_for_abssum_metrics(
    predictions: List[str], labels: List[str], texts: Union[List[str], None] = None
) -> Tuple[List[str], List[str], np.ndarray, np.ndarray, Dict[str, List[str]]]:
    predictions = ["\n".join(sent_tokenize(pred)) for pred in predictions]
    labels = ["\n".join(sent_tokenize(label)) for label in labels]
    # Calculate zero- and uni-word summaries (that must be excluded from calculating
    # ROUGE-1 and ROUGE-L for the former and ROUGE-2 for both)
    stemmer = porter.PorterStemmer()
    tokenized_labels = [tokenize.tokenize(x, stemmer) for x in labels]
    len_tokenized_labels = np.array([len(x) for x in tokenized_labels])
    is_uniword = len_tokenized_labels == 1
    is_zeroword = len_tokenized_labels == 0

    if texts is not None:
        tokenized = {}
        tokenized["preds"] = [tokenize.tokenize(x, stemmer) for x in predictions]
        tokenized["texts"] = [tokenize.tokenize(x, stemmer) for x in texts]
        tokenized["labels"] = tokenized_labels
    else:
        tokenized = None
    return predictions, labels, is_zeroword, is_uniword, tokenized


def calculate_abstractiveness(
    tokenized_preds: List[List[str]],
    tokenized_labels: List[List[str]],
    tokenized_texts: List[List[str]],
) -> Dict[str, float]:
    result = {}
    for use_modified in [False, True]:
        for n in range(1, 5):
            pred_ngram_overlaps = []
            label_ngram_overlaps = []
            for pred, label, text in zip(
                tokenized_preds, tokenized_labels, tokenized_texts
            ):
                pred_pair_ngram_overlap = calculate_ngram_overlap(
                    pred, text, n, use_modified
                )
                if pred_pair_ngram_overlap is not None:
                    pred_ngram_overlaps.append(pred_pair_ngram_overlap)
                label_pair_ngram_overlap = calculate_ngram_overlap(
                    label, text, n, use_modified
                )
                if label_pair_ngram_overlap is not None:
                    label_ngram_overlaps.append(label_pair_ngram_overlap)
            key = f"ngram_overlap_{n}" if use_modified else f"novel_ngrams_{n}"
            result["pred_" + key] = (
                np.mean(pred_ngram_overlaps) if len(pred_ngram_overlaps) > 0 else -1
            )
            result["label_" + key] = (
                np.mean(label_ngram_overlaps) if len(label_ngram_overlaps) > 0 else -1
            )
    return result



from typing import Union
from datasets import Dataset
import numpy as np
from transformers import set_seed

from .al_strategy_utils import assign_ue_scores_for_unlabeled_data, get_X_pool_subsample
from ..utils.transformers_dataset import TransformersDataset


def ensp(
    model,
    X_pool: Union[np.ndarray, Dataset, TransformersDataset],
    n_instances: int,
    **kwargs,
):
    aggregation = kwargs.get("aggregation", "var")
    func_for_agg = getattr(np, aggregation)
    larger_is_less_uncertain = False
    if aggregation in ["mean", "median", "max", "min"]:
        larger_is_less_uncertain = True
    use_log = kwargs.get("use_log", False)
    mc_iterations = kwargs.get("mc_iterations", 1)
    seed = kwargs.get("seed", 42)

    generate_kwargs = dict(to_numpy=True)
    if kwargs.get("enable_dropout", False):
        if seed is not None:
            set_seed(seed)
        # Since BART exploits F.dropout instead of nn.Dropout, we can only turn it on via .train()
        model.model.train()
        update_kwargs = {"do_sample": False, "to_eval_mode": True}
    else:
        model.model.eval()
        update_kwargs = {
            "do_sample": True,
            "to_eval_mode": False,
            "top_p": kwargs.get("generate_top_p", 0.95),
        }
    generate_kwargs.update(update_kwargs)

    if mc_iterations <= 1:
        num_beams = kwargs.get("num_beams", 10)
        mc_iterations = num_beams / model._trainer_kwargs.generation_num_beams
        num_return_sequences = kwargs.get("num_return_sequences", 10)
        assert (
            num_return_sequences <= num_beams
        ), f"`num_beams` must be >= `num_return_sequences`, got values {num_beams}, {num_return_sequences}"

    if kwargs.get("subsample_size_mc_dropout", True):
        X_pool_subsample, subsample_indices = get_X_pool_subsample(
            X_pool, mc_iterations, model.seed
        )
    else:
        X_pool_subsample, subsample_indices = X_pool, np.arange(len(X_pool))

    if isinstance(mc_iterations, int):
        summaries, log_scores = [], []
        for _ in range(mc_iterations):
            generate_output = model.generate(X_pool_subsample, **generate_kwargs)
            log_scores.append(generate_output["sequences_scores"])
            summaries.append(generate_output["predictions"])
    else:
        generate_output = model.generate(
            X_pool_subsample,
            num_beams=num_beams,
            num_return_sequences=num_return_sequences,
            **generate_kwargs,
        )
        log_scores = generate_output["sequences_scores"]
        # summaries = generate_output["predictions"]

    scores = np.r_[log_scores]
    if not use_log:
        scores = np.exp(scores)
    subsample_uncertainty_estimates = func_for_agg(scores, axis=0)
    if larger_is_less_uncertain:
        subsample_uncertainty_estimates = -subsample_uncertainty_estimates

    argsort = np.argsort(-subsample_uncertainty_estimates)
    subsample_query_idx = argsort[:n_instances]
    query = X_pool_subsample.select(subsample_query_idx)
    query_idx = subsample_indices[subsample_query_idx]

    uncertainty_estimates = assign_ue_scores_for_unlabeled_data(
        len(X_pool), subsample_indices, subsample_uncertainty_estimates
    )

    return query_idx, query, uncertainty_estimates



"""
CNN model for text classification and utils
Adapted from https://github.com/chriskhanhtran/CNN-Sentence-Classification-PyTorch
"""
import logging

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from tokenizers import Tokenizer
from tokenizers.models import WordLevel
from tokenizers.normalizers import Lowercase
from tokenizers.pre_tokenizers import Whitespace
from transformers import PreTrainedTokenizerFast

from ..utils.embeddings import load_embeddings
from ..utils.general import DictWithGetattr

log = logging.getLogger()


def init_text_classification_cnn(
    model_cfg, embeddings, word2idx, num_labels, only_tokenizer=False
):
    # create tokenizer
    tokenizer_model = WordLevel(word2idx, "[UNK]")
    tokenizer = Tokenizer(tokenizer_model)
    tokenizer.normalizer = Lowercase()
    tokenizer.pre_tokenizer = Whitespace()
    hf_tokenizer = PreTrainedTokenizerFast(
        tokenizer_object=tokenizer, pad_token="[PAD]", unk_token="[UNK]"
    )
    if only_tokenizer:
        return None, hf_tokenizer
    # build CNN for text classification
    if embeddings is None and model_cfg.embeddings_path is not None:
        embeddings, word2idx = load_embeddings(
            model_cfg.embeddings_path, model_cfg.embeddings_cache_dir
        )
    model = TextClassificationCNN(
        pretrained_embedding=embeddings,
        freeze_embedding=model_cfg.freeze_embedding,
        vocab_size=model_cfg.vocab_size,
        embed_dim=model_cfg.embed_dim,
        filter_sizes=model_cfg.filter_sizes,
        num_filters=model_cfg.num_filters,
        num_classes=num_labels,
        dropout=model_cfg.classifier_dropout,
    )
    return model, hf_tokenizer


class TextClassificationCNN(nn.Module):
    """An 1D Convulational Neural Network for Sentence Classification."""

    def __init__(
        self,
        pretrained_embedding=None,
        freeze_embedding=False,
        vocab_size=None,
        embed_dim=300,
        filter_sizes=[3, 4, 5],
        num_filters=[100, 100, 100],
        num_classes=2,
        dropout=0.5,
    ):
        """
        The constructor for CNN_NLP class.

        Args:
            pretrained_embedding (torch.Tensor): Pretrained embeddings with
                shape (vocab_size, embed_dim)
            freeze_embedding (bool): Set to False to fine-tune pretraiend
                vectors. Default: False
            vocab_size (int): Need to be specified when not pretrained word
                embeddings are not used.
            embed_dim (int): Dimension of word vectors. Need to be specified
                when pretrained word embeddings are not used. Default: 300
            filter_sizes (List[int]): List of filter sizes. Default: [3, 4, 5]
            num_filters (List[int]): List of number of filters, has the same
                length as `filter_sizes`. Default: [100, 100, 100]
            n_classes (int): Number of classes. Default: 2
            dropout (float): Dropout rate. Default: 0.5
        """

        super(TextClassificationCNN, self).__init__()
        # Embedding layer
        if pretrained_embedding is not None:
            self.vocab_size, self.embed_dim = pretrained_embedding.shape
            self.embedding = nn.Embedding.from_pretrained(
                pretrained_embedding,
                padding_idx=self.vocab_size - 1,
                freeze=freeze_embedding,
            )
        else:
            self.embed_dim = embed_dim
            self.embedding = nn.Embedding(
                num_embeddings=vocab_size,
                embedding_dim=self.embed_dim,
                padding_idx=0,
                max_norm=5.0,
            )
        # Conv Network
        self.conv1d_list = nn.ModuleList(
            [
                nn.Conv1d(
                    in_channels=self.embed_dim,
                    out_channels=num_filters[i],
                    kernel_size=filter_sizes[i],
                )
                for i in range(len(filter_sizes))
            ]
        )
        # Fully-connected layer and Dropout
        self.fc = nn.Linear(np.sum(num_filters), num_classes)
        self.dropout = nn.Dropout(p=dropout)
        self.loss_fn = nn.CrossEntropyLoss()
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        # additional param for embeddings model
        self.return_embeddings = False
        self.to(self.device)

    def forward(self, input_ids, labels=None, output_hidden_states=False, **kwargs):
        """Perform a forward pass through the network.

        Args:
            input_ids (torch.Tensor): A tensor of token ids with shape
                (batch_size, max_sent_length)

        Returns:
            logits (torch.Tensor): Output logits with shape (batch_size,
                n_classes)
        """

        # Get embeddings from `input_ids`. Output shape: (b, max_len, embed_dim)
        x_embed = self.embedding(input_ids).float()

        # Permute `x_embed` to match input shape requirement of `nn.Conv1d`.
        # Output shape: (b, embed_dim, max_len)
        x_reshaped = x_embed.permute(0, 2, 1)

        # Apply CNN and ReLU. Output shape: (b, num_filters[i], L_out)
        x_conv_list = [F.relu(conv1d(x_reshaped)) for conv1d in self.conv1d_list]

        # Max pooling. Output shape: (b, num_filters[i], 1)
        x_pool_list = [
            F.max_pool1d(x_conv, kernel_size=x_conv.shape[2]) for x_conv in x_conv_list
        ]

        # Concatenate x_pool_list to feed the fully connected layer.
        # Output shape: (b, sum(num_filters))
        x_fc = torch.cat([x_pool.squeeze(dim=2) for x_pool in x_pool_list], dim=1)

        # Compute logits. Output shape: (b, n_classes)
        x_drop = self.dropout(x_fc)
        logits = self.fc(x_drop)

        loss = None
        if labels is not None:
            loss = self.loss_fn(logits, labels.view(-1))
            results = {"loss": loss, "logits": logits}
        else:
            results = {"logits": logits}
        if output_hidden_states or self.return_embeddings:
            results["last_hidden_state"] = torch.mean(x_embed, dim=1)
            results["hidden_states"] = [torch.mean(x_embed, dim=1)]
        results = DictWithGetattr(results)
        return results

    def save_pretrained(self, save_path):
        torch.save(self.state_dict(), save_path)

    def load_pretrained(self, load_path):
        self.load_state_dict(torch.load(load_path))



from typing import Union

import numpy as np
import torch
from datasets.arrow_dataset import Dataset

from .al_strategy_utils import (
    take_idx,
    calculate_alps_scores,
)
from ..utils.cluster_utils import kmeans
from ..utils.transformers_dataset import TransformersDataset


def alps(
    model,
    X_pool: Union[Dataset, TransformersDataset],
    n_instances: int,
    X_train: Union[Dataset, TransformersDataset],
    **alps_kwargs,
):
    """
    Employs the surprisal embedding of w, which is obtained from the likelihoods of masked tokens from
    pre-trained language models. Doesn't require fine-tuning. https://aclanthology.org/2020.emnlp-main.637/
    """
    logits = model.predict_logits(X_pool)

    kwargs = dict(
        # Necessary
        model_wr=model,
        dataloader_or_data=X_pool,
        # General
        data_is_tokenized=False,
        # data_config=None,
        batch_size=model._batch_size_kwargs.eval_batch_size // 8,
        to_numpy=True,
        logits=logits,
        # train_probas=train_logits,
        # data_is_tokenized=data_is_tokenized,
        tokenizer=model.tokenizer,
        task=model.task,
        text_name=model.data_config["text_name"],
        label_name=model.data_config["label_name"],
    )
    scores_or_vectors = calculate_alps_scores(**kwargs)
    vectors = torch.nn.functional.normalize(scores_or_vectors)
    clustering = kmeans

    query_idx = np.array(clustering(vectors, k=n_instances))

    query = take_idx(X_pool, query_idx)

    # Uncertainty estimates are not defined for ALPS
    uncertainty_estimates = np.zeros(len(X_pool))

    return query_idx, query, uncertainty_estimates



import json
import logging
from collections import defaultdict
from pathlib import Path
from pandas import DataFrame

from datasets import Dataset

from .preprocessing import (
    _add_id_column_to_datasets,
    _use_train_subset,
    _filter_quantiles,
)

log = logging.getLogger()


def load_from_json_or_csv(config, task, cache_dir=None):
    text_name = config.text_name
    label_name = config.label_name
    default_extra_key = "translation" if task == "nmt" else None
    extra_key = config.get("extra_key", default_extra_key)

    path = Path(config.path) / config.dataset_name / "train.csv"
    # The dataset may be nod processed yet
    if path.exists():
        dataset = Dataset.from_csv(str(path))
    else:
        path = Path(config.path) / config.dataset_name / "train.json"
        if path.exists():
            dataset = Dataset.from_json(str(path))
        else:
            convert_list_of_dicts_to_dict(
                Path(config.path) / config.dataset_name, "dataset.json", text_name
            )
            dataset = Dataset.from_json(str(path))

    if extra_key is not None:
        dataset = Dataset.from_pandas(DataFrame(dataset[extra_key]))
    dataset = dataset.remove_columns(
        [x for x in dataset.column_names if x not in [text_name, label_name, "id"]]
    )

    if isinstance(config.get("test_size_split", None), int):
        train_size_split = (len(dataset) - config.get("test_size_split")) / len(dataset)
    else:
        train_size_split = config.get("train_size_split", 0.8)
    splitted_dataset = dataset.train_test_split(
        train_size=train_size_split, shuffle=True, seed=config.get("seed", 42),
    )
    train_dataset, test_dataset = splitted_dataset["train"], splitted_dataset["test"]

    log.info(f"Loaded train size: {len(train_dataset)}")
    log.info(f"Loaded test size: {len(test_dataset)}")
    log.info("Dev dataset coincides with test dataset")

    if getattr(config, "filter_quantiles", None) is not None:
        train_dataset = _filter_quantiles(
            train_dataset,
            config.filter_quantiles,
            cache_dir,
            text_name,
            config.tokenizer_name,
        )

    if getattr(config, "use_subset", None) is not None:
        train_dataset = _use_train_subset(
            train_dataset,
            config.use_subset,
            getattr(config, "seed", 42),
            task,
            label_name,
        )

    if ("id" not in train_dataset.column_names) and config.get("add_id_column", True):
        train_dataset, test_dataset = _add_id_column_to_datasets(
            [train_dataset, test_dataset]
        )

    return [train_dataset, test_dataset, test_dataset, None]


def convert_list_of_dicts_to_dict(path, filename, text_name):
    with open(Path(path) / filename) as f:
        data = json.load(f)
    data_dict = defaultdict(list)
    for item in data:
        for key, value in item.items():
            if isinstance(value, dict):
                value = list(value.values())
            # Since keys may start with '_'
            if (key == text_name or key[1:] == text_name) and isinstance(value, list):
                value = "\n".join(value)
            if key.startswith("_"):
                key = key[1:]
            data_dict[key].append(value)
    dataset = Dataset.from_dict(data_dict)
    dataset.to_json(Path(path) / "train.json")



def get_classifier_dropout_kwargs(
    pretrained_model_name: str, classifier_dropout: float
):
    if "distilbert" in pretrained_model_name:
        key = "seq_classif_dropout"
    elif "deberta" in pretrained_model_name:
        key = "pooler_dropout"
    elif "xlnet" in pretrained_model_name:
        key = "summary_last_dropout"
    elif "distilrubert" in pretrained_model_name:
        key = "dropout"
    elif "rubert-base" in pretrained_model_name:
        key = "hidden_dropout_prob"
    else:
        key = "classifier_dropout"
    return {key: classifier_dropout}


def get_tokenizer_kwargs(pretrained_model_name: str, task: str):
    if task == "ner" and "roberta" in pretrained_model_name:
        return dict(add_prefix_space=True)
    return {}



import logging
import os
from pathlib import Path
from shutil import rmtree
from typing import Union, Tuple, List

import numpy as np
import transformers.utils.logging
from datasets import Dataset, load_dataset, concatenate_datasets
from hydra import compose, initialize_config_dir, core
from omegaconf import OmegaConf
from torch.utils.data import DataLoader

from .calculate_tracin_score import calculate_outlier_scores
from .plasm import label_data, concatenate_data
from ..al4nlp.constructors.construct_wrapper import construct_wrapper
from ..al4nlp.pool_subsampling_strategies.random_subsampling import random_subsampling
from ..al4nlp.utils.general import (
    dill_dump,
    get_target_model_checkpoints,
    add_new_model_to_time_dict,
    json_load,
)
from ..al4nlp.utils.transformers_dataset import TransformersDataset

log = logging.getLogger(__name__)
transformers.utils.logging.disable_progress_bar()


OmegaConf.register_new_resolver(
    "to_string", lambda x: x.replace("/", "_").replace("-", "_"), replace=True
)
OmegaConf.register_new_resolver(
    "get_patience_value", lambda dev_size: 1000 if dev_size == 0 else 5, replace=True
)
DEFAULT_CONFIG_NAME = "cls_plasm"


def pipeline_plasm(
    data_dir: Union[str, Path],
    annotation_or_annotation_dir: Union[str, Path, List[List[int]], np.ndarray],
    config_name: str = None,
    config_dir: str = None,
    train_model: bool = True,
) -> Tuple[Union[Dataset, TransformersDataset], "ModelWrapper"]:
    # Initialize config
    core.global_hydra.GlobalHydra.instance().clear()
    default_config_dir = os.path.join(os.getcwd(), "../jupyterlab_demo/configs/")
    initialize_config_dir(config_dir=config_dir or default_config_dir)
    config = compose(config_name=config_name or DEFAULT_CONFIG_NAME)
    # Load data
    log.info("Starting post processing")
    data_dir = Path(data_dir)
    labeled_data = load_dataset(
        "json", data_files=str(data_dir / "labeled.json"), split="train"
    )
    unlabeled_data = load_dataset(
        "json", data_files=str(data_dir / "unlabeled.json"), split="train"
    )
    # Load annotation
    if isinstance(annotation_or_annotation_dir, (list, np.ndarray)):
        annotation = annotation_or_annotation_dir
    else:
        annotation = json_load(Path(annotation_or_annotation_dir) / "annotation.json")
    # Load label2id
    tags = json_load(data_dir / "tags.json")
    id2label = {idx: tag for idx, tag in enumerate(tags)}
    label2id = {v: k for k, v in id2label.items()}
    labeled_idx = [i for i, inst_ann in enumerate(annotation) if inst_ann is not None]
    # Remove Nones and transform labels to ids
    annotation = [
        list(map(lambda tag: label2id[tag], annotation[idx])) for idx in labeled_idx
    ]
    annotated_data = unlabeled_data.select(labeled_idx).add_column(
        config.data.label_name, annotation
    )
    labeled_data = concatenate_datasets(
        [labeled_data, annotated_data], info=labeled_data.info
    )
    ### Specify pseudo-labeled data peculiarities
    # Whether to use label_smoothing for pseudo labeled data;
    # True -> use classic LS, 'natural' -> use returned probas, None -> do not use
    pl_label_smoothing = config.post_processing.get("label_smoothing")
    # If we want to increase the weight of the "real labeled data" compared to pseudo labeled data
    labeled_weight = config.post_processing.get("labeled_weight", 1)
    # Uncertainty threshold
    unc_threshold = config.post_processing.get("uncertainty_threshold", None)
    # We can either filter by absolute value or by a quantile
    unc_filt_by_quant = unc_threshold == "adaptive" or config.post_processing.get(
        "filter_by_quantile", None
    )
    # Whether pseudo-labeling will be performed over a subsample of the data
    use_subsample_for_pl = config.post_processing.get("use_subsample_for_pl", False)
    # Whether to use TracIn
    use_tracin = config.post_processing.tracin.get("use", False)
    # Whether to use TracIn
    tracin_quantile = config.post_processing.tracin.get("quantile", "adaptive")
    log.info(
        f"Pseudo-labeling:\n"
        + f"PL label smoothing factor: {pl_label_smoothing}\n"
        + f"Labeled data weight: {labeled_weight}\n"
        + f"Using TracIn: {use_tracin}\n"
        + (f"TracIn quantile: {tracin_quantile}" if use_tracin else "")
        + f"Uncertainty threshold: {unc_threshold}\n"
        + (
            f"Uncertainty filter by quantile: {unc_filt_by_quant}\n"
            if unc_threshold is not None
            else ""
        )
        + f"PL using subsample: {use_subsample_for_pl}\n"
    )
    # Construct and train a pseudo-labeling model
    labeling_wrapper = construct_wrapper(
        config,
        model_cfg=config.labeling_model,
        dev_data=None,
        labels_or_id2label=id2label,
        name="labeling",
    )
    log.info("Labeling wrapper constructed. Started fitting...")
    labeling_wrapper.fit(labeled_data, from_scratch=True)
    log.info("Labeling wrapper fitted.")
    # Setting adaptive parameters if necessary
    lab_model_quality = getattr(labeling_wrapper, "best_metric", None)
    if lab_model_quality is None:
        if tracin_quantile == "adaptive":
            log.warning(
                "You set TracIn filtering share to be adaptive but no dev data was used for evaluation. "
                + "Using value 0.1."
            )
            tracin_quantile = 0.1
        if pl_label_smoothing == "adaptive":
            log.warning(
                "You set PL label smoothing coef. to be adaptive but no dev data was used for evaluation. "
                + "Disabling PL label smoothing."
            )
            pl_label_smoothing = None
        if unc_threshold == "adaptive":
            log.warning(
                "You set uncertainty threshold coef. to be adaptive but no dev data was used for evaluation. "
                + "Disabling uncertainty threshold filtering."
            )
            unc_threshold = None

    # Set adaptive coefs
    if pl_label_smoothing == "adaptive":
        pl_label_smoothing = lab_model_quality
        log.info(f"Using pseudo-labeling parameter {pl_label_smoothing:.5f}.")
    if unc_threshold == "adaptive":
        if unc_filt_by_quant:
            unc_threshold = 1 - lab_model_quality
        else:
            unc_threshold = lab_model_quality
        log.info(f"Using uncertainty parameter {unc_threshold:.5f}.")
    if tracin_quantile == "adaptive":
        tracin_quantile = 1 - lab_model_quality
        log.info(f"Using TracIn filtering coef {tracin_quantile:.5f}.")
    # Set data for pseudo-labeling
    if use_subsample_for_pl:
        if isinstance(use_subsample_for_pl, float):
            use_subsample_for_pl = round(len(unlabeled_data) * use_subsample_for_pl)
        subsample_idx = random_subsampling(
            np.ones(len(unlabeled_data)), use_subsample_for_pl
        )
        unlabeled_data = unlabeled_data.select(subsample_idx)
    # Pseudo labeling!
    log.info("Starting pseudo-labeling...")
    pseudo_labeled_data, filtered_data_share = label_data(
        unlabeled_data,
        labeling_wrapper,
        threshold=unc_threshold,
        uncertainty_filter_by_quantile=unc_filt_by_quant,
        pl_label_smoothing=pl_label_smoothing,
        id2label=id2label,
    )
    log.info("Pseudo-labeling done.")
    # Concatenate the labeled-by-oracle and pseudo-labeled data
    successor_train_data = concatenate_data(
        labeled_data, pseudo_labeled_data, labeled_weight
    )
    if not train_model:
        return successor_train_data

    # Train the successor model
    successor_wrapper = construct_wrapper(
        config,
        model_cfg=config.successor_model,
        dev_data=None,
        labels_or_id2label=id2label,
        name="successor",
    )
    log.info("Successor wrapper constructed. Started fitting...")
    successor_wrapper.fit(successor_train_data)
    log.info("Successor wrapper fitted.")
    tracin_quantile -= filtered_data_share
    if not use_tracin or tracin_quantile <= 0:
        return successor_train_data, successor_wrapper

    tokenizer = successor_wrapper.tokenizer
    tokenized_data = successor_wrapper.tokenize_data(
        tokenizer=tokenizer,
        data=pseudo_labeled_data,
        text_name=successor_wrapper.data_config["text_name"],
        label_name=successor_wrapper.data_config["label_name"],
    )

    framework = config.framework
    framework_is_flair = framework == "flair"
    if framework_is_flair:
        from flair.datasets import DataLoader as FlairDataLoader

        dataloader = FlairDataLoader(tokenized_data, batch_size=1,)
    else:
        dataloader = DataLoader(tokenized_data, batch_size=1, pin_memory=True,)

    tracin_path = data_dir / "tracin"
    tracin_path.mkdir(exist_ok=True)
    model_path = tracin_path / f"tmp_successor_model_{config.seed}"
    model_weights_paths = get_target_model_checkpoints(config, framework=framework)
    dataloader_path = tracin_path / "tmp_dataloader"

    if framework_is_flair:
        # flair models don't work with dill
        successor_wrapper.model.save(model_path)
    else:
        dill_dump(successor_wrapper.model, model_path)

    dill_dump(dataloader, dataloader_path)
    # Calculate TracIn score
    idx_to_tag = (
        successor_wrapper.data_config["idx_to_tag"] if framework_is_flair else None
    )
    log.info("TracIn started")
    scores, *_ = calculate_outlier_scores(
        model_path=model_path,
        weights_paths=model_weights_paths,
        dataloader_path=dataloader_path,
        work_dir=tracin_path,
        max_num_processes=config.post_processing.tracin.max_num_processes,
        task=successor_wrapper.task,
        nu=config.post_processing.tracin.nu,
        framework=framework,
        idx_to_tag=idx_to_tag,
        config=config,
    )
    log.info("TracIn finished")
    # Change fp16 dict
    successor_wrapper._trainer_kwargs["fp16"] = successor_wrapper._trainer_kwargs.get(
        "final_model_fp16", False
    )
    # Modify the name to avoid overwriting time-file
    successor_wrapper.name = (
        successor_wrapper.name + f"_tracin_quantile_{np.round(tracin_quantile, 4)}"
    )
    # Add initial empty fit/predict lists of the model to the time dict
    add_new_model_to_time_dict(successor_wrapper.time_dict_path, successor_wrapper.name)
    # Load key metrics of the successor model (f1-score for NER, accuracy for CLS)
    outliers = np.argwhere(scores > np.quantile(scores, (1 - tracin_quantile))).ravel()
    pseudo_labeled_data_to_use = [
        pseudo_labeled_data[i]
        for i in range(len(pseudo_labeled_data))
        if i not in outliers
    ]

    if successor_wrapper.task == "ner":
        tokenization_kwargs = {"is_split_into_words": True}
    elif successor_wrapper.task == "cls":
        tokenization_kwargs = {}
    else:
        raise NotImplementedError

    pseudo_labeled_data_to_use = TransformersDataset(
        pseudo_labeled_data_to_use,
        text_column_name=config.data.text_name,
        label_column_name=config.data.label_name,
        tokenization_kwargs=tokenization_kwargs,
        task=pseudo_labeled_data.task,
        id2label=pseudo_labeled_data.id2label,
        label_smoothing=pseudo_labeled_data.label_smoothing,
    )
    labeled_weight = successor_wrapper.model_config.training.get("labeled_weight", 1.0)
    filtered_data = concatenate_data(
        labeled_data, pseudo_labeled_data_to_use, labeled_weight
    )
    log.info("Successor fitting...")
    successor_wrapper.fit(filtered_data)
    log.info("Successor fitted.")
    rmtree(successor_wrapper._trainer_kwargs["serialization_dir"])
    os.remove(model_path)
    os.remove(dataloader_path)
    return filtered_data, successor_wrapper



from typing import Union

import numpy as np
from datasets.arrow_dataset import Dataset

from .al_strategy_utils import (
    take_idx,
    _best_versus_second_best,
)
from ..utils.transformers_dataset import TransformersDataset


def breaking_ties(
    model,
    X_pool: Union[np.ndarray, Dataset, TransformersDataset],
    n_instances: int,
    **kwargs,
):
    """
    Selects instances which have a small margin between their most likely and second
    most likely predicted class.
    https://www.researchgate.net/publication/220321064_Active_Learning_to_Recognize_Multiple_Types_of_Plankton
    """
    probas = model.predict_proba(X_pool)
    uncertainty_estimates = np.apply_along_axis(
        lambda x: _best_versus_second_best(x), 1, probas
    )

    argsort = np.argsort(uncertainty_estimates)
    query_idx = argsort[:n_instances]
    query = take_idx(X_pool, query_idx)

    return query_idx, query, uncertainty_estimates
